# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Collections
from collections import OrderedDict

# Filesystem
from pathlib import Path
from os.path import splitext

# GEM
from geode_gem.engine.utils import (are_equivalent_timestamps,
                                    copy,
                                    generate_identifier,
                                    get_data,
                                    get_boot_datetime_as_timestamp)

from geode_gem.engine.game import Game
from geode_gem.engine.console import Console
from geode_gem.engine.emulator import Emulator

from geode_gem.engine.lib.database import Database
from geode_gem.engine.lib.configuration import Configuration

# Logging
import logging

from logging.config import fileConfig

# System
from os import getpid
from os import name as os_name
from sys import exit as sys_exit


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GEM(object):

    Instance = "gem"
    Version = "0.11.2"

    Log = "gem.log"
    Logger = "log.conf"
    Consoles = "consoles.conf"
    Emulators = "emulators.conf"
    Databases = "databases.conf"
    Environment = "environment.conf"

    def __init__(self, config, local, debug=False):
        """ Constructor

        Parameters
        ----------
        config : pathlib.Path or str
            Default config folder
        local : pathlib.Path or str
            Default data folder
        debug : bool, optional
            Debug mode status (default: False)
        """

        if isinstance(config, str):
            config = Path(config)

        if isinstance(local, str):
            local = Path(local)

        if type(debug) is not bool:
            debug = False

        # ----------------------------------------
        #   Variables
        # ----------------------------------------

        # Debug mode
        self.debug = debug

        # Migration mode
        self.__need_migration = False

        # Data list
        self.__data = dict(
            consoles=dict(),
            emulators=dict(),
            environment=list()
        )

        # Rename list
        self.__rename = OrderedDict()

        # Configurations
        self.__configurations = dict(
            consoles=None,
            emulators=None,
            environment=None
        )

        # Process identifier
        self.__pid = int()
        self.__lock = False

        # ----------------------------------------
        #   Initialize filepaths
        # ----------------------------------------

        self.__config = config.joinpath(GEM.Instance).expanduser()
        self.__local = local.joinpath(GEM.Instance).expanduser()

        self.__log_path = self.__local.joinpath(f"{GEM.Instance}.log")
        self.__lock_path = self.__local.joinpath(".lock")

        self.__backup_path = self.__local.joinpath(f"backup.{GEM.Instance}.db")
        self.__database_path = self.__local.joinpath(f"{GEM.Instance}.db")

        for folder in (self.__config, self.__local):
            if not folder.exists():
                folder.mkdir(mode=0o755, parents=True)

        # ----------------------------------------
        #   Initialize objects
        # ----------------------------------------

        # Initialize lock file
        self.__lock = self.__init_lock()

        # Avoid to initialize a new instance if an existing one is present
        if not self.__lock:

            # Initialize logging module
            self.__init_logger()

            # Initialize sqlite database
            self.__init_database()

            self.logger.debug(f"Set local folder as {self.__local}")
            self.logger.debug(f"Set config folder as {self.__config}")

    def __init_lock(self):
        """ Initialize lock file

        Create a lock file which avoid to access to the database with multiple
        instance simultaneous

        The .lock file is stored into the XDG_DATA_HOME/gem directory.

        Inside, we store these values: "GEM_PID BOOT_TIMESTAMP".
        """

        if not os_name == "posix":
            logging.getLogger(self.Instance).warning(
                "Your operating system do not support POSIX standard")
            return False

        proc_path = Path("/proc")
        if not proc_path.exists():
            logging.getLogger(self.Instance).warning(
                "Your operating system do not support process file system")
            return False

        # Retrieve current boot session timestamp
        try:
            timestamp = get_boot_datetime_as_timestamp()
            if timestamp is None:
                return False

        except FileNotFoundError:
            return False

        # A lock file already exists on file system
        if self.__lock_path.exists():
            # Ensure to always retrieve the first line which contains PID
            with self.__lock_path.open('r') as pipe:
                content = pipe.read().split()

            if content:
                self.__pid, *more = content[0:]

                # Check if specified PID still alive
                if proc_path.joinpath(self.__pid).exists():

                    # No way to check if this PID is really owned by GEM
                    if not more or not more[0]:
                        return True

                    if are_equivalent_timestamps(timestamp, more[0], delta=4):
                        return True

        # Generate a new lock file
        self.__pid = getpid()

        with self.__lock_path.open('wb') as pipe:
            # Register current PID into file
            pipe.write(bytes(f"{self.__pid} {timestamp}", "UTF-8"))

            try:
                from fcntl import flock, LOCK_EX, LOCK_NB

                # Perform the lock operation on file
                flock(pipe, LOCK_EX | LOCK_NB)

            except ImportError:
                pass

        return False

    def __init_logger(self):
        """ Initialize logger

        Create a logger object based on logging library
        """

        if self.__log_path.exists():
            copy(self.__log_path, str(self.__log_path) + ".old")

        # Define log path with a global variable
        logging.log_path = str(self.__log_path)

        # Generate logger from log.conf
        fileConfig(str(get_data("data", "config", GEM.Logger)))

        self.logger = logging.getLogger("gem")
        if not self.debug:
            self.logger.setLevel(logging.INFO)

    def __init_database(self):
        """ Initialize database

        Check GEM database from local folder and update if needed columns and
        data
        """

        try:
            config = Configuration(
                get_data("data", "config", GEM.Databases), strict=False)

            # Check GEM database file
            self.database = Database(
                self.__database_path, config, self.logger)

            # Check current GEM version
            version = self.database.select("gem", "version")

            # Check Database inner version and GEM version
            if not version == GEM.Version:
                if version is None:
                    self.logger.info("Generate a new database")
                    version = GEM.Version

                else:
                    self.logger.info(f"Update database to v.{GEM.Version}")

                self.database.modify("gem",
                                     {"version": GEM.Version},
                                     {"version": version})

            else:
                self.logger.debug(f"Use GEM API v.{GEM.Version}")

            # Check integrity and migrate if necessary
            self.logger.info("Check database integrity")
            if not self.database.check_integrity():
                self.logger.warning("Database need a migration")
                self.__need_migration = True

            else:
                self.logger.info("Current database is up-to-date")
                self.__need_migration = False

        except OSError as error:
            self.logger.exception(f"Cannot access to database: {error}")
            sys_exit(error)

        except ValueError as error:
            self.logger.exception(f"A wrong value occur: {error}")
            sys_exit(error)

        except Exception as error:
            self.logger.exception(f"An error occur: {error}")
            sys_exit(error)

    def __init_configurations(self):
        """ Initalize configuration

        Check consoles.conf and emulators.conf from user config folder and copy
        default one if not exists
        """

        if not self.__config.exists():
            self.logger.debug(f"Generate {self.__config} folder")

            self.__config.mkdir(mode=0o755, parents=True)

        # Check GEM configuration files
        for filename in (GEM.Consoles, GEM.Emulators):
            path = Path(self.get_config(filename))
            if path.exists():
                self.logger.debug(f"Read {path} configuration file")

                # Store Configuration object
                self.__configurations[path.stem] = \
                    Configuration(path, strict=False)

        path = Path(self.get_config(GEM.Environment))

        self.logger.debug(f"Read {path} configuration file")

        self.__configurations[path.stem] = Configuration(path)

    def __init_emulators(self):
        """ Initalize emulators

        Load emulators.conf from user config folder and generate Emulator
        objects from data
        """

        self.__data["emulators"].clear()

        emulators = self.__configurations["emulators"]

        if emulators is not None:
            for section in emulators.sections():
                self.add_emulator(section, emulators.items(section))

            self.logger.debug(
                f"{len(self.emulators)} emulator(s) has been founded")

    def __init_consoles(self):
        """ Initalize consoles

        Load consoles.conf from user config folder and generate Console objects
        from data
        """

        self.__data["consoles"].clear()

        consoles = self.__configurations["consoles"]

        if consoles is not None:
            for section in consoles.sections():
                self.add_console(section, consoles.items(section))

            self.logger.debug(
                f"{len(self.consoles)} console(s) has been founded")

    def init(self):
        """ Initalize data from configuration files

        This function allow to reset API by reloading default configuration
        files
        """

        if self.__need_migration:
            raise RuntimeError("GEM database need a migration")

        # Check if default configuration file exists
        self.__init_configurations()

        # Load user emulators
        self.__init_emulators()
        # Load user consoles
        self.__init_consoles()

    def check_database(self):
        """ Check database and migrate to lastest GEM version if needed
        """

        if self.__need_migration:
            self.logger.info("Backup database")

            # Database backup
            copy(self.__database_path, self.__backup_path)

            # Remove previous database
            self.__database_path.unlink()

            # ----------------------------------------
            #   Initialize new database
            # ----------------------------------------

            try:
                config = Configuration(
                    get_data("data", "config", GEM.Databases))

                previous_database = Database(
                    self.__backup_path, config, self.logger)

                new_database = Database(
                    self.__database_path, config, self.logger)

                new_database.insert("gem", {"version": GEM.Version})

                # ----------------------------------------
                #   Migrate data from previous database
                # ----------------------------------------

                self.logger.info("Start database migration")

                # ----------------------------------------
                #   Migrate game by game
                # ----------------------------------------

                games = previous_database.select("games", ['*'])

                if games is not None:

                    # Get current table columns
                    old_columns_name = previous_database.get_columns("games")

                    # Get new table columns
                    new_columns_name = new_database.get_columns("games")

                    for counter, row in enumerate(games):
                        counter += 1

                        row_data = dict()

                        for element in row:
                            column = old_columns_name[row.index(element)]

                            # Avoid to retrieve columns which are no more used
                            if column in new_columns_name:
                                row_data[column] = element

                        new_database.insert("games", row_data)

                        yield counter, len(games)

                # ----------------------------------------
                #   Remove backup
                # ----------------------------------------

                self.logger.info("Migration complete")
                self.__need_migration = False

                del previous_database
                del self.database

                setattr(self, "database", new_database)

            except Exception as error:
                self.logger.exception(
                    f"An error occurs during migration: {error}")

                self.logger.info("Restore database backup")

                copy(self.__backup_path, self.__database_path)

            # Remove backup
            self.__backup_path.unlink()

        yield None

    def write_object(self, data):
        """ Write data into a specific configuration file

        Parameters
        ----------
        data : object
            Data structure to save

        Returns
        -------
        bool
            return True if object was successfully writed, False otherwise
        """

        config = None

        if isinstance(data, Console):
            config = self.__configurations["consoles"]

        elif isinstance(data, Emulator):
            config = self.__configurations["emulators"]

        if config is not None:
            structure = data.as_dict()

            for key, value in structure.items():
                if value is None:
                    value = str()

                if type(value) is bool:

                    if value:
                        value = "yes"

                    else:
                        value = "no"

                if type(value) is Emulator:
                    value = value.id

                config.modify(data.name, key, value)

            config.update()

    def write_data(self, *files):
        """ Write data into configuration files and database

        Returns
        -------
        bool
            return True if files were successfully writed, False otherwise

        Notes
        -----
        Previous files are backup
        """

        self.logger.debug("Store GEM data into disk")

        # ----------------------------------------
        #   Configuration files
        # ----------------------------------------

        try:
            # Check GEM configuration files
            for path in files:

                # Get configuration filename for storage
                name, ext = splitext(path)

                # Backup configuration file
                if self.get_config(path).exists():
                    self.logger.debug(f"Backup {path} file")

                    copy(self.get_config(path), self.get_config(f"~{path}"))

                    self.get_config(path).unlink()

                # Create a new configuration object
                config = Configuration(self.get_config(path))

                # Feed configuration with new data
                for element in sorted(self.__data[name]):
                    structure = self.__data[name][element].as_dict()

                    for key, value in sorted(structure.items()):
                        if value is None or value == "None":
                            value = str()

                        if type(value) is bool:
                            if value:
                                value = "yes"
                            else:
                                value = "no"

                        elif type(value) is Emulator:
                            value = value.id

                        config.modify(
                            self.__data[name][element].name, key, value)

                # Write new configuration file
                self.logger.info(f"Write configuration into {path} file")
                config.update()

        except Exception as error:
            self.logger.exception(f"Cannot write configuration: {error}")
            return False

        # ----------------------------------------
        #   Database file
        # ----------------------------------------

        try:
            for previous, emulator in self.__rename.items():

                # Update games which use a renamed emulator
                self.database.update("games",
                                     {"emulator": emulator.id},
                                     {"emulator": previous})

                self.logger.info(f"Update old {previous} references from "
                                 f"database to {emulator.id}")

        except Exception as error:
            self.logger.exception(f"Cannot write database: {error}")
            return False

        return True

    def get_config(self, *args):
        """ Retrieve configuration data

        Parameters
        ----------
        args : str, optional
            Optional path
        """

        return self.__config.joinpath(*args).expanduser()

    def get_local(self, *args):
        """ Retrieve local data

        Parameters
        ----------
        args : str, optional
            Optional path
        """

        return self.__local.joinpath(*args).expanduser()

    def is_locked(self):
        """ Check if database is locked

        Returns
        -------
        bool
            Lock status
        """

        return self.__lock

    def free_lock(self):
        """ Remove lock file if present
        """

        if self.__lock_path.exists():
            self.__lock_path.unlink()

    @property
    def pid(self):
        """ Return application process identifier

        Returns
        -------
        int
            Process identifier
        """

        return self.__pid

    @property
    def log(self):
        """ Return application log filepath

        Returns
        -------
        pathlib.Path
            Log file instance
        """

        return self.__log_path

    @property
    def emulators(self):
        """ Return emulators dict

        Returns
        -------
        dict
            emulators dictionary with identifier as keys
        """

        return self.__data["emulators"]

    def get_emulators(self):
        """ Return emulators list

        Returns
        -------
        list
            Emulators list
        """

        return list(self.__data["emulators"].values())

    def get_emulator(self, emulator):
        """ Get a specific emulator

        Parameters
        ----------
        emulator : str
            Emulator identifier or name

        Returns
        -------
        Emulator or None
            Found emulator
        """

        if emulator is not None and len(emulator) > 0:

            if emulator in self.__data["emulators"].keys():
                return self.__data["emulators"].get(emulator, None)

            # Check if emulator use name instead of identifier
            identifier = generate_identifier(emulator)

            if identifier in self.__data["emulators"].keys():
                return self.__data["emulators"].get(identifier, None)

        return None

    def add_emulator(self, name, informations):
        """ Add a new emulator

        Parameters
        ----------
        name : str
            Emulator name
        informations : dict
            Emulator information as dictionary

        Returns
        -------
        gem.engine.emulator.Emulator
            New emulator object
        """

        data = dict(name=name)

        for option, value in informations:
            convert_keys = dict(
                save="savestates",
                snaps="screenshots")

            if option in convert_keys.keys():
                option = convert_keys[option]

            if value is not None:
                data[option] = value

        emulator = Emulator(**data)

        self.__data["emulators"][emulator.id] = emulator

        return emulator

    def update_emulator(self, emulator):
        """ Update a specific emulator

        Parameters
        ----------
        emulator : gem.engine.emulator.Emulator
            Emulator instance
        """

        if emulator is not None:
            self.__data["emulators"][emulator.id] = emulator

    def delete_emulator(self, emulator):
        """ Delete a specific emulator

        Parameters
        ----------
        emulator : str
            Emulator identifier

        Raises
        ------
        IndexError
            if emulator not exists
        """

        if emulator not in self.__data["emulators"].keys():
            raise IndexError(f"Cannot access to {emulator} in emulators list")

        del self.__data["emulators"][emulator]

    def rename_emulator(self, previous, identifier):
        """ Rename an emulator and all associate objects (consoles and games)

        Parameters
        ----------
        previous : str
            Emulator previous identifier
        identifier : str
            Emulator new identifier

        Raises
        ------
        IndexError
            if emulator not exists
        """

        # Avoid to rename an emulator with the same name :D
        if not previous == identifier:

            if identifier not in self.__data["emulators"].keys():
                raise IndexError(
                    f"Cannot access to {identifier} in emulators list")

            # Retrieve emulator object
            self.__rename[previous] = self.__data["emulators"][identifier]

            # Update consoles which use previous emulator
            for console_identifier in self.__data["consoles"].keys():
                console = self.__data["consoles"][console_identifier]

                if console is not None and console.emulator.id == previous:
                    console.emulator = self.__rename[previous]

    @property
    def consoles(self):
        """ Return consoles dict

        Returns
        -------
        dict
            Consoles dictionary with identifier as keys
        """

        return self.__data["consoles"]

    def get_consoles(self):
        """ Return consoles list

        Returns
        -------
        list
            Consoles list
        """

        return list(self.__data["consoles"].values())

    def get_console(self, console):
        """ Get a specific console

        Parameters
        ----------
        console : str
            Console identifier or name

        Returns
        -------
        gem.engine.api.Console or None
            Found console

        Examples
        --------
        >>> g = GEM()
        >>> g.init()
        >>> g.get_console("nintendo-nes")
        <gem.engine.api.Console object at 0x7f174a986b00>
        """

        if console is not None and len(console) > 0:

            if console in self.__data["consoles"].keys():
                return self.__data["consoles"].get(console, None)

            # Check if console use name instead of identifier
            identifier = generate_identifier(console)

            if identifier in self.__data["consoles"].keys():
                return self.__data["consoles"].get(identifier, None)

        return None

    def add_console(self, name, informations):
        """ Add a new console

        Parameters
        ----------
        name : str
            Console name
        informations : dict
            Console information as dictionary

        Returns
        -------
        gem.engine.console.Console
            New console object
        """

        data = dict(name=name)

        for option, value in informations:
            convert_keys = dict(
                roms="path",
                exts="extensions",
                save="savestates",
                snaps="screenshots")

            if option in convert_keys.keys():
                option = convert_keys[option]

            if value is not None:
                data[option] = value

        console = Console(self, **data)

        self.__data["consoles"][console.id] = console

        return console

    def update_console(self, console):
        """ Update a specific console

        Parameters
        ----------
        console : gem.engine.console.Console
            Console instance
        """

        if console is not None:
            self.__data["consoles"][console.id] = console

    def delete_console(self, console):
        """ Delete a specific console

        Parameters
        ----------
        console : str
            Console identifier

        Raises
        ------
        IndexError
            if console not exists
        """

        if console not in self.__data["consoles"].keys():
            raise IndexError(f"Cannot access to {console} in consoles list")

        del self.__data["consoles"][console]

    @property
    def environment(self):
        """ Return environment dict

        Returns
        -------
        dict
            environment dictionary
        """

        return self.__configurations["environment"]

    def get_games(self):
        """ List all games from register consoles

        Returns
        -------
        list
            Games list
        """

        games = list()

        for identifier, console in self.consoles.items():
            games.extend(console.get_games())

        return games

    def get_game(self, console, game):
        """ Get game from a specific console

        Parameters
        ----------
        console : str
            Console identifier
        game : str
            Game identifier

        Returns
        -------
        gem.engine.api.Game or None
            Game object

        Raises
        ------
        IndexError
            if console not exists

        Examples
        --------
        >>> g = GEM()
        >>> g.init()
        >>> g.get_game("nintendo-nes", "gremlins-2-the-new-batch-usa")
        <gem.engine.api.Game object at 0x7f174a986f60>
        """

        if console not in self.__data["consoles"]:
            raise IndexError(f"Cannot access to {console} in consoles list")

        # Check console games list
        return self.__data["consoles"][console].get_game(game)

    def get_game_tags(self):
        """ Retrieve avaialable game tags from database

        Returns
        -------
        list
            Tags list
        """

        result = self.database.select("games", "tags")

        if result is not None:
            tags = list()

            for tag in result:
                tags.extend(tag.split(';'))

            return sorted(list(set(tags)))

        return list()

    def update_game(self, game):
        """ Update a game in database

        Parameters
        ----------
        game : gem.engine.api.Game
            Game object

        Returns
        -------
        bool
            return True if update successfully, False otherwise

        Raises
        ------
        TypeError
            if game type is not gem.engine.api.Game
        """

        if type(game) is not Game:
            raise TypeError(
                "Wrong type for game, expected gem.engine.api.Game")

        # Store game data
        data = game.as_dict()

        # Translate value as string for database
        for key, value in data.items():

            if value is None:
                data[key] = str()

            elif type(value) is bool:
                data[key] = str(int(value))

            elif type(value) is int:
                data[key] = str(value)

            elif type(value) is Emulator:
                data[key] = str(value.name)

        # Update game in database
        self.logger.debug(f"Update {game.name} database entry")

        self.database.modify("games", data, {"filename": game.path.name})

        # Update game environment variables
        self.logger.debug(f"Update {game.name} environment variables")

        self.environment.remove_section(game.id)

        if len(game.environment) > 0:
            self.environment.add_section(game.id)

            for key, value in game.environment.items():
                self.environment.set(game.id, key.upper(), value)

        self.environment.update()

    def delete_game(self, game):
        """ Delete a specific game

        Parameters
        ----------
        game : gem.engine.api.Game
            Game object

        Raises
        -------
        TypeError
            if game type is not gem.api.Game
        """

        if type(game) is not Game:
            raise TypeError(
                "Wrong type for game, expected gem.engine.api.Game")

        results = self.database.get("games", {"filename": game.path.name})

        if results is not None and len(results) > 0:
            self.logger.info(f"Remove {game.name} from database")

            self.database.remove("games", {"filename": game.path.name})

        # Update game environment variables
        self.logger.debug(f"Remove {game.name} environment variables")

        self.environment.remove_section(game.id)

        self.environment.update()



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Datetime
from datetime import datetime

# GObject
try:
    from gi.repository import GObject

except ImportError as error:
    from sys import exit

    exit("Cannot found python3-gobject module: %s" % str(error))

# Processus
from subprocess import PIPE
from subprocess import Popen

# System
from os import environ

# Threading
from threading import Thread

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GamesLoadingThread(Thread, GObject.GObject):

    __gsignals__ = {
        "games-terminate": (GObject.SignalFlags.RUN_LAST, None, [object]),
    }

    def __init__(self, parent, console):
        """ Constructor

        Parameters
        ----------
        parent : geode_gem.ui.interface.MainWindow
            Main interface to access public variables
        console : geode_gem.engine.console.Console
            Console object instance
        """

        Thread.__init__(self)
        GObject.GObject.__init__(self)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.parent = parent
        self.logger = parent.logger

        self.name = console.id
        self.console = console

        self.error_message = None

    def run(self):
        """ Launch GamesLoadingThread instance
        """

        self.logger.info(f"Init games for {self.console.name}")

        # Load games list if the game directory exists
        try:
            self.console.init_games(reload_files=False)

        except FileNotFoundError:
            self.error_message = \
                _("Cannot retrieve console path '%s'") % self.console.path

        except NotADirectoryError:
            self.error_message = \
                _("Console path '%' is not a directory") % self.console.path

        except PermissionError:
            self.error_message = \
                _("Cannot read console path '%s'") % self.console.path

        # Call game-terminate signal on main window
        self.parent.emit("games-terminate", self)


class GameThread(Thread, GObject.GObject):

    __gsignals__ = {
        "game-started": (GObject.SignalFlags.RUN_FIRST, None, [object]),
        "game-terminate": (GObject.SignalFlags.RUN_LAST, None, [object]),
    }

    def __init__(self, parent, game, fullscreen=False):
        """ Constructor

        Parameters
        ----------
        parent : geode_gem.ui.interface.MainWindow
            Main interface to access public variables
        game : geode_gem.engine.game.Game
            Game object instance
        fullscreen : bool
            Fullscreen mode
        """

        Thread.__init__(self)
        GObject.GObject.__init__(self)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.parent = parent
        self.logger = parent.logger

        self.name = game.id
        self.game = game

        self.fullscreen = fullscreen

        self.delta = None
        self.error = False

        # ------------------------------------
        #   Generate data
        # ------------------------------------

        self.path = parent.api.get_local("logs", game.id + ".log")

    def run(self):
        """ Launch GameThread instance

        This function start a new processus with Popen and wait until it stop.
        When it finish, GameThread emit a signal to main interface.
        """

        # Call game-started signal on main window
        self.parent.emit("game-started", self.game)

        started = datetime.now()

        self.logger.info("Launch %s" % self.game.name)

        try:
            command = self.game.command(self.fullscreen)

            self.logger.debug("Command: %s" % ' '.join(command))

            # ------------------------------------
            #   Check environment
            # ------------------------------------

            # Get a copy of current environment
            environment = environ.copy()

            # Check if current game has specific environment variable
            for key, value in self.game.environment.items():
                environment[key] = value

            # ------------------------------------
            #   Start process
            # ------------------------------------

            self.logger.info("Log to %s" % self.path)

            # Logging process output
            with open(self.path, 'w') as pipe:
                self.proc = Popen(
                    command,
                    stdin=PIPE,
                    stdout=pipe,
                    stderr=pipe,
                    env=environment,
                    start_new_session=True,
                    universal_newlines=True)

                self.proc.communicate()

            self.logger.info("Close %s" % self.game.name)

            self.proc.terminate()

            # ------------------------------------
            #   Play time
            # ------------------------------------

            self.delta = (datetime.now() - started)

        except OSError as error:
            self.logger.error("Cannot access to game: %s" % str(error))
            self.error = True

        except MemoryError as error:
            self.logger.error("A memory error occur: %s" % str(error))
            self.error = True

        except KeyboardInterrupt:
            self.logger.info("Terminate by keyboard interrupt")

        except Exception as error:
            self.logger.info("An exception error occur: %s" % str(error))
            self.error = True

        # Call game-terminate signal on main window
        self.parent.emit("game-terminate", self)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.engine.utils import parse_timedelta

from geode_gem.ui.data import Icons
from geode_gem.ui.utils import replace_for_markup
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import GdkPixbuf, Gtk, Pango

# System
from os import environ

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class ParametersDialog(CommonWindow):

    def __init__(self, parent, game):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        game : gem.engine.game.Game
            Game object
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              _("Game properties"),
                              Icons.Symbolic.GAMING,
                              classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.api = parent.api

        self.game = game

        self.help_data = {
            "order": [
                _("Description"),
                _("Parameters"),
            ],
            _("Description"): [
                _("Emulator default arguments can use custom parameters to "
                    "facilitate file detection."),
                _("Nintendo console titles are identified by a 6 character "
                    "identifier known as a GameID. This GameID is only used "
                    "with some emulators like Dolphin-emu. For more "
                    "informations, consult emulators documentation."),
                _("Tags are split by commas.")
            ],
            _("Parameters"): {
                "<key>": _("Use game key"),
                "<name>": _("Use ROM filename"),
                "<lname>": _("Use ROM lowercase filename"),
                "<rom_path>": _("Use ROM folder path"),
                "<rom_file>": _("Use ROM file path"),
                "<conf_path>": _("Use emulator configuration file path"),
            }
        }

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, -1)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Stack
        # ------------------------------------

        self.stack = Gtk.Stack()

        self.sidebar_stack = Gtk.StackSwitcher()

        # Properties
        self.stack.set_margin_top(12)
        self.stack.set_transition_type(Gtk.StackTransitionType.NONE)

        self.sidebar_stack.set_margin_top(12)
        self.sidebar_stack.set_stack(self.stack)
        self.sidebar_stack.set_halign(Gtk.Align.CENTER)

        # ------------------------------------
        #   Grids
        # ------------------------------------

        self.grid_content = Gtk.Box()

        grid_parameters = Gtk.Box()

        grid_environment = Gtk.Box()
        grid_environment_buttons = Gtk.Box()

        grid_statistic = Gtk.Box()
        grid_statistic_total = Gtk.Box()
        grid_statistic_average = Gtk.Box()

        # Properties
        self.grid_content.set_spacing(6)
        self.grid_content.set_orientation(Gtk.Orientation.VERTICAL)

        grid_parameters.set_spacing(6)
        grid_parameters.set_homogeneous(False)
        grid_parameters.set_orientation(Gtk.Orientation.VERTICAL)

        grid_environment.set_spacing(12)
        grid_environment.set_homogeneous(False)

        Gtk.StyleContext.add_class(
            grid_environment_buttons.get_style_context(), "linked")
        grid_environment_buttons.set_spacing(-1)
        grid_environment_buttons.set_orientation(Gtk.Orientation.VERTICAL)

        grid_statistic.set_spacing(6)
        grid_statistic.set_homogeneous(False)
        grid_statistic.set_orientation(Gtk.Orientation.VERTICAL)

        grid_statistic_total.set_spacing(12)
        grid_statistic_total.set_homogeneous(True)

        grid_statistic_average.set_spacing(12)
        grid_statistic_average.set_homogeneous(True)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % (
                replace_for_markup(self.game.name)))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   Emulators
        # ------------------------------------

        label_emulator = Gtk.Label()

        self.model = Gtk.ListStore(GdkPixbuf.Pixbuf, str)
        self.combo = Gtk.ComboBox()

        cell_icon = Gtk.CellRendererPixbuf()
        cell_name = Gtk.CellRendererText()

        # Properties
        label_emulator.set_use_markup(True)
        label_emulator.set_halign(Gtk.Align.CENTER)
        label_emulator.set_markup("<b>%s</b>" % _("Alternative emulator"))

        self.model.set_sort_column_id(1, Gtk.SortType.ASCENDING)

        self.combo.set_model(self.model)
        self.combo.set_id_column(1)
        self.combo.pack_start(cell_icon, False)
        self.combo.add_attribute(cell_icon, "pixbuf", 0)
        self.combo.pack_start(cell_name, True)
        self.combo.add_attribute(cell_name, "text", 1)

        cell_icon.set_padding(4, 0)

        # ------------------------------------
        #   Arguments
        # ------------------------------------

        self.entry_arguments = Gtk.Entry()

        # Properties
        self.entry_arguments.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, Icons.Symbolic.TERMINAL)
        self.entry_arguments.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)

        # ------------------------------------
        #   Key
        # ------------------------------------

        label_key = Gtk.Label()

        self.entry_key = Gtk.Entry()

        # Properties
        label_key.set_margin_top(12)
        label_key.set_halign(Gtk.Align.CENTER)
        label_key.set_use_markup(True)
        label_key.set_markup("<b>%s</b>" % _("GameID"))

        self.entry_key.set_placeholder_text(_("No default value"))
        self.entry_key.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, Icons.Symbolic.PASSWORD)
        self.entry_key.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)

        # ------------------------------------
        #   Tags
        # ------------------------------------

        label_tags = Gtk.Label()

        self.entry_tags = Gtk.Entry()
        self.completion_tags = Gtk.EntryCompletion()

        self.store_tags = Gtk.ListStore(str)

        # Properties
        label_tags.set_margin_top(12)
        label_tags.set_halign(Gtk.Align.CENTER)
        label_tags.set_use_markup(True)
        label_tags.set_markup("<b>%s</b>" % _("Tags"))

        self.entry_tags.set_tooltip_text(
            _("Use comma to separate tags"))
        self.entry_tags.set_placeholder_text(
            _("Use comma to separate tags"))
        self.entry_tags.set_completion(self.completion_tags)
        self.entry_tags.set_icon_from_icon_name(
            Gtk.EntryIconPosition.PRIMARY, Icons.Symbolic.ADD_TEXT)
        self.entry_tags.set_icon_from_icon_name(
            Gtk.EntryIconPosition.SECONDARY, Icons.Symbolic.CLEAR)

        self.completion_tags.set_model(self.store_tags)
        self.completion_tags.set_popup_single_match(True)
        self.completion_tags.set_popup_completion(True)
        self.completion_tags.set_text_column(0)
        self.completion_tags.set_match_func(self.__on_entry_match_tag)

        # ------------------------------------
        #   Statistics
        # ------------------------------------

        scroll_statistic = Gtk.ScrolledWindow()
        viewport_statistic = Gtk.Viewport()

        label_statistic_total = Gtk.Label()
        self.label_statistic_total_value = Gtk.Label()

        label_statistic_average = Gtk.Label()
        self.label_statistic_average_value = Gtk.Label()

        # Properties
        label_statistic_total.set_label(_("Total play time"))
        label_statistic_total.set_halign(Gtk.Align.END)
        label_statistic_total.set_valign(Gtk.Align.CENTER)
        label_statistic_total.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)

        self.label_statistic_total_value.set_label(_("No data"))
        self.label_statistic_total_value.set_halign(Gtk.Align.START)
        self.label_statistic_total_value.set_valign(Gtk.Align.CENTER)
        self.label_statistic_total_value.set_use_markup(True)

        label_statistic_average.set_label(_("Average play time"))
        label_statistic_average.set_halign(Gtk.Align.END)
        label_statistic_average.set_valign(Gtk.Align.CENTER)
        label_statistic_average.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)

        self.label_statistic_average_value.set_label(_("No data"))
        self.label_statistic_average_value.set_halign(Gtk.Align.START)
        self.label_statistic_average_value.set_valign(Gtk.Align.CENTER)
        self.label_statistic_average_value.set_use_markup(True)

        # ------------------------------------
        #   Environment variables
        # ------------------------------------

        scroll_environment = Gtk.ScrolledWindow()
        viewport_environment = Gtk.Viewport()

        image_environment_add = Gtk.Image()
        self.button_environment_add = Gtk.Button()

        image_environment_remove = Gtk.Image()
        self.button_environment_remove = Gtk.Button()

        self.store_environment_keys = Gtk.ListStore(str)

        self.store_environment = Gtk.ListStore(str, str)
        self.treeview_environment = Gtk.TreeView()

        treeview_column_environment = Gtk.TreeViewColumn()

        self.treeview_cell_environment_key = Gtk.CellRendererCombo()
        self.treeview_cell_environment_value = Gtk.CellRendererText()

        # Properties
        scroll_environment.set_shadow_type(Gtk.ShadowType.OUT)

        image_environment_add.set_from_icon_name(
            Icons.Symbolic.ADD, Gtk.IconSize.BUTTON)
        image_environment_remove.set_from_icon_name(
            Icons.Symbolic.REMOVE, Gtk.IconSize.BUTTON)

        self.treeview_environment.set_model(self.store_environment)
        self.treeview_environment.set_headers_clickable(False)
        self.treeview_environment.set_headers_visible(False)

        treeview_column_environment.pack_start(
            self.treeview_cell_environment_key, True)
        treeview_column_environment.pack_start(
            self.treeview_cell_environment_value, True)

        treeview_column_environment.add_attribute(
            self.treeview_cell_environment_key, "text", 0)
        treeview_column_environment.add_attribute(
            self.treeview_cell_environment_value, "text", 1)

        self.treeview_cell_environment_key.set_padding(12, 6)
        self.treeview_cell_environment_key.set_property("text-column", 0)
        self.treeview_cell_environment_key.set_property("editable", True)
        self.treeview_cell_environment_key.set_property("has-entry", True)
        self.treeview_cell_environment_key.set_property(
            "model", self.store_environment_keys)
        self.treeview_cell_environment_key.set_property(
            "placeholder_text", _("Key"))
        self.treeview_cell_environment_key.set_property(
            "ellipsize", Pango.EllipsizeMode.END)

        self.treeview_cell_environment_value.set_padding(12, 6)
        self.treeview_cell_environment_value.set_property("editable", True)
        self.treeview_cell_environment_value.set_property(
            "placeholder_text", _("Value"))
        self.treeview_cell_environment_value.set_property(
            "ellipsize", Pango.EllipsizeMode.END)

        self.treeview_environment.append_column(
            treeview_column_environment)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.stack.add_titled(grid_parameters, "parameters", _("Parameters"))

        grid_parameters.pack_start(label_emulator, False, False, 0)
        grid_parameters.pack_start(self.combo, False, False, 0)
        grid_parameters.pack_start(self.entry_arguments, False, False, 0)
        grid_parameters.pack_start(label_key, False, False, 0)
        grid_parameters.pack_start(self.entry_key, False, False, 0)
        grid_parameters.pack_start(label_tags, False, False, 0)
        grid_parameters.pack_start(self.entry_tags, False, False, 0)

        self.stack.add_titled(scroll_statistic, "statistic", _("Statistic"))

        scroll_statistic.add(viewport_statistic)
        viewport_statistic.add(grid_statistic)

        grid_statistic.pack_start(grid_statistic_total, False, False, 0)
        grid_statistic.pack_start(grid_statistic_average, False, False, 0)

        grid_statistic_total.pack_start(
            label_statistic_total, True, True, 0)
        grid_statistic_total.pack_start(
            self.label_statistic_total_value, True, True, 0)

        grid_statistic_average.pack_start(
            label_statistic_average, True, True, 0)
        grid_statistic_average.pack_start(
            self.label_statistic_average_value, True, True, 0)

        self.stack.add_titled(
            grid_environment, "environment", _("Environment"))

        scroll_environment.add(viewport_environment)
        viewport_environment.add(self.treeview_environment)

        self.button_environment_add.add(image_environment_add)
        self.button_environment_remove.add(image_environment_remove)

        grid_environment_buttons.pack_start(
            self.button_environment_add, False, False, 0)
        grid_environment_buttons.pack_start(
            self.button_environment_remove, False, False, 0)

        grid_environment.pack_start(grid_environment_buttons, False, False, 0)
        grid_environment.pack_start(scroll_environment, True, True, 0)

        self.grid_content.pack_start(self.sidebar_stack, False, False, 0)
        self.grid_content.pack_start(self.stack, True, True, 0)

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.grid_content)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        self.combo.connect(
            "changed", self.__on_selected_emulator)

        self.entry_arguments.connect(
            "icon-press", self.__on_entry_clear)
        self.entry_tags.connect(
            "icon-press", self.__on_entry_clear)
        self.entry_key.connect(
            "icon-press", self.__on_entry_clear)

        self.completion_tags.connect(
            "match-selected", self.__on_entry_write_tag)

        self.treeview_cell_environment_key.connect(
            "edited", self.__on_edited_cell)
        self.treeview_cell_environment_value.connect(
            "edited", self.__on_edited_cell)

        self.button_environment_add.connect(
            "clicked", self.__on_append_item)
        self.button_environment_remove.connect(
            "clicked", self.__on_remove_item)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("Close"), Gtk.ResponseType.CLOSE)
        self.add_button(_("Accept"), Gtk.ResponseType.APPLY, Gtk.Align.END)

        self.add_help(self.help_data)

        for emulator in self.main_parent.api.emulators.values():

            if emulator.exists:

                icon = self.main_parent.get_pixbuf_from_cache(
                    "emulators", 22, emulator.id, emulator.icon)

                if icon is None:
                    icon = self.main_parent.icons_blank.get(22)

                row = self.model.append([icon, emulator.name])

                if self.game.emulator is not None \
                   and emulator.name == self.game.emulator.name:
                    self.combo.set_active_iter(row)

        self.combo.set_wrap_width(int(len(self.model) / 6))

        # Emulator parameters
        if len(self.game.default) > 0:
            self.entry_arguments.set_text(self.game.default)

        # Game ID
        if self.game.key is not None:
            self.entry_key.set_text(self.game.key)

        # Game tags
        if len(self.game.tags) > 0:
            self.entry_tags.set_text(', '.join(self.game.tags))

        for tag in self.main_parent.api.get_game_tags():
            self.store_tags.append([tag])

        # Game statistics
        if not parse_timedelta(self.game.play_time) == "00:00:00":
            self.label_statistic_total_value.set_label(
                parse_timedelta(self.game.play_time))

            if self.game.played > 0:
                self.label_statistic_average_value.set_label(
                    parse_timedelta(self.game.play_time / self.game.played))

        # Environment variables
        for key in sorted(self.game.environment.keys()):
            self.store_environment.append([key, self.game.environment[key]])

        for key in sorted(environ.copy().keys()):
            self.store_environment_keys.append([key])

        self.combo.grab_focus()

    def __on_selected_emulator(self, widget=None):
        """ Select an emulator in combobox and update parameters placeholder

        Other Parameters
        ----------------
        widget : Gtk.Widget
            Object which receive signal (Default: None)
        """

        emulator = \
            self.main_parent.api.get_emulator(self.combo.get_active_id())

        self.entry_arguments.set_placeholder_text(_("No default value"))
        self.set_response_sensitive(Gtk.ResponseType.APPLY, False)

        if emulator is not None:

            if len(emulator.default) > 0:
                self.entry_arguments.set_placeholder_text(emulator.default)

            # Allow to validate dialog if selected emulator binary exist
            if emulator.exists:
                self.set_response_sensitive(Gtk.ResponseType.APPLY, True)

    def __on_edited_cell(self, widget, path, text):
        """ Update treerow when a cell has been edited

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        path : str
            Path identifying the edited cell
        text : str
            New text
        """

        if widget == self.treeview_cell_environment_key:
            self.store_environment[path][0] = str(text)

        elif widget == self.treeview_cell_environment_value:
            self.store_environment[path][1] = str(text)

    def __on_append_item(self, widget):
        """ Append a new row in treeview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        self.store_environment.append([str(), str()])

    def __on_remove_item(self, widget):
        """ Remove a row in treeview

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        model, treeiter = \
            self.treeview_environment.get_selection().get_selected()

        if treeiter is not None:
            self.store_environment.remove(treeiter)

    def __on_entry_clear(self, widget, pos, event):
        """ Reset an entry widget when secondary icon is clicked

        Parameters
        ----------
        widget : Gtk.Entry
            Entry widget
        pos : Gtk.EntryIconPosition
            Position of the clicked icon
        event : Gdk.EventButton or Gdk.EventKey
            Event which triggered this signal

        Returns
        -------
        bool
            Function state
        """

        if type(widget) is not Gtk.Entry:
            return False

        if pos == Gtk.EntryIconPosition.SECONDARY \
           and len(widget.get_text()) > 0:
            widget.set_text(str())

            return True

        return False

    def __on_entry_match_tag(self, widget, key, treeiter):
        """ Check if current entry match an entry from completion

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        key : str
            String to match
        treeiter : Gtk.Treeiter
            Row to match

        Returns
        -------
        bool
            Match item status
        """

        text = self.entry_tags.get_text()

        # Retrieve tag from model
        tag = self.store_tags.get_value(treeiter, 0)

        if ',' in text:

            # Retrieve current cursor position
            position = self.entry_tags.get_position()

            # Retrieve commas position
            left_comma = text.rfind(',', 0, position)
            right_comma = text.find(',', position)

            if left_comma == -1 and right_comma == -1:
                text = text

            elif left_comma == -1:
                text = text[:right_comma]

            elif right_comma == -1:
                text = text[left_comma + 1:]

            else:
                text = text[left_comma + 1:right_comma]

        # Avoid to retrieve all the tag when the specified text is empty
        if len(text.strip()) == 0:
            return False

        # Avoid to show the tag when the user already complete it
        if text.strip() == tag:
            return False

        # Check if the tag start with the specified text
        return tag.startswith(text.strip())

    def __on_entry_write_tag(self, widget, model, treeiter):
        """ Write the specified completion item to entry

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        model : Gtk.TreeModel
            Model which contains data
        treeiter : Gtk.Treeiter
            Selected entry from model

        Returns
        -------
        bool
            Handled signal
        """

        text = self.entry_tags.get_text()

        # Retrieve tag from model
        tag = model.get_value(treeiter, 0)

        if len(text) > 0:

            # Retrieve current cursor position
            position = self.entry_tags.get_position()

            # Retrieve commas position
            left_comma = text.rfind(',', 0, position)
            right_comma = text.find(',', position)

            left_text = str()
            if not left_comma == -1:
                left_text = text[:left_comma + 1] + ' '

            right_text = str()
            if not right_comma == -1:
                right_text = text[right_comma:]

            self.entry_tags.set_text(left_text + tag + right_text)
            self.entry_tags.set_position(len(left_text + tag))

        return True



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon
from geode_gem.widgets.menu import GeodeGtkMenu

# GObject
from gi.repository import Gtk, Pango


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonButton(GeodeGtkCommon):

    def __init__(self, subclass, label, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        subclass : Gtk.Button
            Subclass widget type
        label : str
            String use as button label
        """

        GeodeGtkCommon.__init__(self, subclass, **kwargs)

        # Inner widgets
        self.image = None
        # Button image icon name
        self.icon_name = kwargs.get("icon_name", None)

        # ------------------------------------
        #   Properties
        # ------------------------------------

        if self.icon_name is None:
            self.set_label(label)

        else:
            self.set_tooltip_text(label)

            self.image = Gtk.Image.new_from_icon_name(self.icon_name,
                                                      Gtk.IconSize.BUTTON)
            setattr(self.image, "identifier", f"{self.identifier}_image")

        # ------------------------------------
        #   Packing
        # ------------------------------------

        if self.image is not None:
            self.append_widget(self.image)
            self.add(self.image)


class GeodeGtkButton(CommonButton, Gtk.Button):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.button.CommonButton
        """

        CommonButton.__init__(self, Gtk.Button, *args, **kwargs)


class GeodeGtkFileChooserButton(GeodeGtkCommon, Gtk.FileChooserButton):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.button.CommonButton
        """

        GeodeGtkCommon.__init__(self, Gtk.FileChooserButton, **kwargs)


class GeodeGtkFontButton(GeodeGtkCommon, Gtk.FontButton):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.button.CommonButton
        """

        GeodeGtkCommon.__init__(self, Gtk.FontButton, **kwargs)

        # HACK: Set an ellipsize mode for the label inside FontButton
        if kwargs.get("use_ellipsize", False):
            for child in self.get_child():
                if type(child) == Gtk.Label:
                    child.set_ellipsize(Pango.EllipsizeMode.END)


class GeodeGtkMenuButton(CommonButton, Gtk.MenuButton):

    __setters__ = {
        "set_use_popover": False,
    }

    def __init__(self, label, *args, **kwargs):
        """ See geode_gem.ui.widgets.button.CommonButton
        """

        CommonButton.__init__(self, Gtk.MenuButton, label, *args, **kwargs)

        # Inner widgets
        self.submenu = None

        # Properties
        if args:
            self.submenu = GeodeGtkMenu(*args)
            self.append_widget(self.submenu)

        # Packing
        if self.submenu is not None:
            self.set_popup(self.submenu)
            self.submenu.show_all()


class GeodeGtkSpinButton(GeodeGtkCommon, Gtk.SpinButton):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.button.CommonButton
        """

        GeodeGtkCommon.__init__(self, Gtk.SpinButton, **kwargs)


class GeodeGtkToggleButton(CommonButton, Gtk.ToggleButton):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.button.CommonButton
        """

        CommonButton.__init__(self, Gtk.ToggleButton, *args, **kwargs)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GObject
from gi.repository import Gtk, Pango


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class ListBoxItem(Gtk.ListBoxRow):

    def __init__(self):
        """ Constructor
        """

        Gtk.ListBoxRow.__init__(self)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.__widget = None

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        # ------------------------------------
        #   Grids
        # ------------------------------------

        self.grid = Gtk.Box()
        self.grid_labels = Gtk.Box()

        # Properties
        self.grid.set_homogeneous(False)
        self.grid.set_border_width(6)
        self.grid.set_spacing(12)

        self.grid_labels.set_orientation(Gtk.Orientation.VERTICAL)
        self.grid_labels.set_homogeneous(False)
        self.grid_labels.set_spacing(2)

        # ------------------------------------
        #   Labels
        # ------------------------------------

        self.label_title = Gtk.Label()

        self.label_description = Gtk.Label()

        # Properties
        self.label_title.set_line_wrap(True)
        self.label_title.set_halign(Gtk.Align.START)
        self.label_title.set_justify(Gtk.Justification.FILL)
        self.label_title.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)

        self.label_description.set_hexpand(True)
        self.label_description.set_line_wrap(True)
        self.label_description.set_use_markup(True)
        self.label_description.set_no_show_all(True)
        self.label_description.set_halign(Gtk.Align.START)
        self.label_description.set_valign(Gtk.Align.START)
        self.label_description.set_justify(Gtk.Justification.FILL)
        self.label_description.set_line_wrap_mode(Pango.WrapMode.WORD_CHAR)
        self.label_description.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL)

    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        self.add(self.grid)

        self.grid.pack_start(self.grid_labels, True, True, 0)

        self.grid_labels.pack_start(self.label_title, True, True, 0)
        self.grid_labels.pack_start(self.label_description, True, True, 0)

    @staticmethod
    def new(text):
        """ Generate a new ListBoxItem with a specific label

        Parameters
        ----------
        text : str
            Label text
        """

        row = ListBoxItem()
        row.set_option_label(text)

        return row

    def set_option_label(self, text):
        """ Set the option label text

        Parameters
        ----------
        text : str
            Label text
        """

        self.label_title.set_text(text)

    def get_option_label(self):
        """ Retrieve the option label text

        Returns
        -------
        str
            Label text
        """

        return self.label_title.get_text()

    def set_description_label(self, text):
        """ Set the description label text

        Parameters
        ----------
        text : str
            Label text
        """

        if len(text) > 0:
            self.label_description.set_markup(
                "<span size=\"small\">%s</span>" % text)

            self.label_title.set_valign(Gtk.Align.END)

        else:
            self.label_title.set_valign(Gtk.Align.CENTER)

        self.label_description.set_visible(len(text) > 0)

    def set_widget(self, widget):
        """ Set a new internal widget

        Parameters
        ----------
        widget : Gtk.Widget
            Internal widget to set
        """

        # Remove previous widget
        if self.__widget is not None:
            self.grid.remove(self.__widget)

        # Add new widget
        if widget is not None:
            widget.set_valign(Gtk.Align.CENTER)

            if type(widget) in (Gtk.Switch, Gtk.Button, Gtk.SpinButton,
                                Gtk.Image, Gtk.MenuButton):
                widget.set_halign(Gtk.Align.END)
                widget.set_hexpand(False)

                self.grid.set_homogeneous(False)

                self.grid.pack_start(widget, False, False, 0)

            else:
                widget.set_halign(Gtk.Align.FILL)
                widget.set_hexpand(True)

                self.grid.set_homogeneous(True)

                self.grid.pack_start(widget, True, True, 0)

        self.__widget = widget

    def get_widget(self):
        """ Retrieve internal widget

        Returns
        -------
        Gtk.Widget
            Internal widget
        """

        return self.__widget


class ScrolledListBox(Gtk.ScrolledWindow):

    def __init__(self):
        """ Constructor
        """

        Gtk.ScrolledWindow.__init__(self)

        # ----------------------------------------
        #   Prepare interface
        # ----------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_propagate_natural_height(True)

        # ----------------------------------------
        #   Games list
        # ----------------------------------------

        self.listbox = Gtk.ListBox()

        # Properties
        self.listbox.set_activate_on_single_click(True)
        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)

    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        self.add(self.listbox)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from os import access, R_OK
from pathlib import Path

# GEM
from geode_gem.engine.utils import generate_extension, generate_identifier
from geode_gem.engine.game import Game
from geode_gem.engine.emulator import Emulator

# Regex
from re import IGNORECASE
from re import compile as re_compile
from re import match as re_match


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Console(object):

    attributes = {
        "id": str,
        "name": str,
        "icon": Path,
        "path": Path,
        "emulator": Emulator,
        "ignores": list,
        "extensions": list,
        "favorite": bool,
        "recursive": bool
    }

    def __init__(self, parent, **kwargs):
        """ Constructor

        Parameters
        ----------
        parent : gem.engine.api.GEM
            API instance
        """

        # ----------------------------------------
        #   Variables
        # ----------------------------------------

        self.__parent = parent

        self.__games = list()

        self.__files_cache = list()

        # ----------------------------------------
        #   Initialization
        # ----------------------------------------

        # Initialize variables
        self.__init_keys(**kwargs)

    def __init_keys(self, **kwargs):
        """ Initialize object attributes
        """

        for key, key_type in self.attributes.items():

            if key in kwargs.keys():
                value = kwargs[key]

            elif key_type is Emulator or key_type is Path:
                value = None

            elif key_type is bool:
                value = False

            else:
                value = key_type()

            setattr(self, key, value)

            if key_type is Path and type(value) is str:
                if "<local>" in value and hasattr(self.__parent, "get_local"):
                    value = value.replace(
                        "<local>", str(self.__parent.get_local()))

                path = None
                if len(value) > 0 and not value == "None":
                    path = Path(value).expanduser()

                setattr(self, key, path)

            elif key_type is Emulator and type(value) is str:
                setattr(self, key, self.__parent.get_emulator(value))

            elif key_type is bool:

                if value == "yes":
                    setattr(self, key, True)

                elif value == "no":
                    setattr(self, key, False)

            elif key_type is list and type(value) is str:
                setattr(self, key, key_type())

                if len(value.strip()) > 0:
                    value = list(set(value.strip().split(';')))
                    value.sort()

                    setattr(self, key, value)

        if not self.name and self.path:
            setattr(self, "name", self.path.stem)

        setattr(self, "id", generate_identifier(self.name))

        if not self.name:
            setattr(self, "name", self.id)

    def __len__(self):
        """ Retrieve console filtered files number

        Returns
        -------
        int
            Filtered files number
        """

        return len(self.__files_cache)

    def as_dict(self):
        """ Return object as dictionary structure

        Returns
        -------
        dict
            Data structure
        """

        return {
            "icon": str(self.icon),
            "roms": str(self.path),
            "exts": ';'.join(self.extensions),
            "ignores": ';'.join(self.ignores),
            "emulator": self.emulator,
            "favorite": self.favorite,
            "recursive": self.recursive
        }

    def init_games(self, reload_files=True):
        """ Initialize games list from path directory

        Parameters
        ----------
        reload_files : bool
            Force to read path directory again to retrieve games files
            (Default: True)

        Raises
        ------
        FileNotFoundError
            when path directory was not founded
        NotADirectoryError
            when path is not a directory
        PermissionError
            when path did not have read access
        """

        if reload_files:
            self.list_files()

        # Rest games list
        self.__games.clear()

        # Retrieve files from games directory
        for filename in self.__files_cache:
            self.add_game(filename)

    def add_game(self, filename):
        """ Add a new game

        Parameters
        ----------
        filename : str or pathlib.Path
            Game filepath

        Returns
        -------
        gem.engine.game.Game
            Game instance

        Raises
        ------
        ValueError
            when the filename was already added to Console collection
        """

        game = Game(self.__parent, filename)

        if self.get_game(game.id):
            raise ValueError(f"The Game ID '{game.id}' already exists in "
                             f"console '{self.name}'")

        if game.emulator is None:
            game.emulator = self.emulator

        self.__games.append(game)

        if filename not in self.__files_cache:
            self.__files_cache.append(filename)

        return game

    def delete_game(self, game):
        """ Remove a game from console storage

        Parameters
        ----------
        game : gem.engine.game.Game
            Game instance
        """

        if not isinstance(game, Game):
            raise TypeError(f"Cannot use specified game parameter, must be a "
                            f"{repr(Game)} object")

        if game not in self.__games:
            raise KeyError(
                f"Cannot remove game '{game.name}' from '{self.id}' console")

        self.__games.remove(game)

        if game.path in self.__files_cache:
            self.__files_cache.remove(game.path)

    def get_games(self):
        """ Retrieve games list

        Returns
        -------
        list
            Games list
        """

        return self.__games

    def get_game(self, key):
        """ Return specific game from current console

        Parameters
        ----------
        key : str
            Game identifier key

        Returns
        -------
        gem.engine.game.Game or None
            Game instance if found, None otherwise
        """

        return next((game for game in self.__games if game.id == key), None)

    def list_files(self):
        """ List files from console path directory

        Raises
        ------
        ValueError
            when path directory was not specified
        FileNotFoundError
            when path directory was not founded
        NotADirectoryError
            when path is not a directory
        PermissionError
            when path did not have read access
        """

        if self.path is None:
            raise ValueError("Cannot list files without a directory path")

        if not self.path.exists():
            raise FileNotFoundError(
                f"Cannot found '{self.path}' in filesystem")

        elif not self.path.is_dir():
            raise NotADirectoryError(
                f"'{self.path}' is not a directory")

        elif not access(self.path, R_OK):
            raise PermissionError(
                f"Read permission not available for '{self.path}'")

        self.__files_cache.clear()
        for extension in self.extensions:
            files = self.retrieve_files(f"*.{generate_extension(extension)}")

            self.__files_cache.extend([filename for filename in files
                                       if self.path_is_not_ignored(filename)])

        self.__files_cache = sorted(self.__files_cache, key=lambda p: p.stem)

    def path_is_not_ignored(self, path):
        """ Check if game is not ignored by console ignores patterns list

        Parameters
        ----------
        path : pathlib.Path
            File path to check

        Returns
        -------
        bool
            True if path is not ignored, False otherwise
        """

        return all([re_match(element, path.name, IGNORECASE) is None
                    for element in self.ignores])

    def retrieve_files(self, pattern):
        """ Retrieve files list from a specific pattern

        Parameters
        ----------
        pattern : str
            Pattern used to retrieve files with glob method

        Return
        ------
        generator
            Files list as generator object
        """

        if self.recursive:
            return self.path.rglob(pattern)

        return self.path.glob(pattern)

    def search_game(self, key):
        """ Search games from a specific key

        Parameters
        ----------
        key : str
            Key to search in games list (based on identifier and name)

        Returns
        -------
        list
            Found games list
        """

        regex = re_compile(key, IGNORECASE)

        return list(game for game in self.__games
                    if regex.search(game.name) or regex.search(game.id))



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons

# GObject
from gi.repository import Gdk, Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonWindow(object):

    def __init__(self, parent, title, icon, classic=False):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Dialog title
        icon : str
            Default icon name
        classic : bool, optional
            Using classic theme (Default: False)
        """

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.parent = parent
        self.title = title
        self.icon = icon

        self.use_classic_theme = classic

        self.has_help_button = False

        self.help_data = list()
        self.sensitive_data = dict()

        # Ensure to retrieve the root parent
        self.main_parent = getattr(parent, "main_parent", parent)
        # Retrieve logging instance from root parent
        self.logger = self.main_parent.logger

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init signals
        self.__init_signals()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.window = Gtk.Dialog(use_header_bar=not self.use_classic_theme)

        # Properties
        self.window.set_border_width(0)
        self.window.set_can_focus(True)
        self.window.set_default_icon_name(self.icon)
        self.window.set_destroy_with_parent(True)
        self.window.set_keep_above(True)
        self.window.set_modal(True)
        self.window.set_position(Gtk.WindowPosition.CENTER)
        self.window.set_resizable(False)
        self.window.set_title(self.title)
        self.window.set_transient_for(self.main_parent)
        self.window.set_type_hint(Gdk.WindowTypeHint.DIALOG)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid = self.window.get_content_area()

        # Remove the old action grid
        self.grid.remove(self.grid.get_children()[0])

        self.grid_tools = Gtk.Box()
        self.grid_actions = Gtk.Box()
        self.grid_actions_buttons = Gtk.ButtonBox()

        # Properties
        self.grid.set_spacing(18)
        self.grid.set_border_width(18)

        self.grid_actions_buttons.set_spacing(12)
        self.grid_actions_buttons.set_margin_top(6)
        self.grid_actions_buttons.set_layout(Gtk.ButtonBoxStyle.END)

        # ------------------------------------
        #   Headerbar
        # ------------------------------------

        if not self.use_classic_theme:

            self.headerbar = self.window.get_header_bar()

            self.headerbar_image = Gtk.Image()

            # Properties
            self.headerbar.set_show_close_button(True)

            self.headerbar_image.set_from_icon_name(
                self.icon, Gtk.IconSize.LARGE_TOOLBAR)

        # ------------------------------------
        #   Help button
        # ------------------------------------

        self.image_help = Gtk.Image()

        self.button_help = Gtk.Button()

        # Properties
        self.image_help.set_from_icon_name(
            Icons.Symbolic.HELP, Gtk.IconSize.MENU)

        self.button_help.set_image(self.image_help)
        self.button_help.set_relief(Gtk.ReliefStyle.NONE)
        self.button_help.set_tooltip_text(_("Get some help about this dialog"))

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.grid.pack_start(self.grid_tools, False, False, 0)

        self.grid.pack_end(self.grid_actions, False, False, 0)

        self.grid_actions.pack_end(self.grid_actions_buttons, False, False, 0)

        if not self.use_classic_theme:
            self.headerbar.pack_start(self.headerbar_image)

        # Gtk.Window
        if self.parent is None:
            self.window.add(self.grid)

    def __init_signals(self):
        """ Initialize widgets signals
        """

        # Gtk.Window
        if self.parent is None:
            self.window.connect("delete-event", self.destroy)

    def __on_show_help(self, widget):
        """ Launch help dialog

        Parameters
        ----------
        widget : Gtk.Widget
            Object which receive signal
        """

        dialog = HelpDialog(self.parent,
                            _("Help"),
                            '\n\n'.join(self.help_data),
                            Icons.Symbolic.HELP)

        dialog.set_size(640, 480)

        dialog.run()
        dialog.destroy()

    def __on_execute_method_from_widget(self, widget, method):
        """ Execute specified method from widget if available

        Parameters
        ----------
        widget : Gtk.Widget
            Widget which received the event
        method : str
            Method to execute
        """

        if hasattr(widget, method):
            getattr(widget, method)()

    def show_all(self):
        """ Recursively shows a widget, and any child widgets
        """

        if len(self.grid_tools.get_children()) == 0:
            self.grid.remove(self.grid_tools)
        else:
            self.grid_tools.set_spacing(6)

        if len(self.grid_actions_buttons.get_children()) == 0:
            self.grid.remove(self.grid_actions)
        else:
            self.grid_actions.set_spacing(12)

        self.window.hide()
        self.window.unrealize()

        self.window.show_all()

    def run(self):
        """ Start dialog

        Returns
        -------
        Gtk.ResponseType
            Dialog response
        """

        self.show_all()

        return self.window.run()

    def destroy(self, *args):
        """ Destroy dialog
        """

        self.window.destroy()

    def add_widget(self, widget, align=Gtk.Align.START,
                   expand=False, fill=False, padding=int()):
        """ Add a widget to dialog headerbar

        Parameters
        ----------
        widget : Gtk.Widget
            Widget to add
        align : Gtk.Align, optional
            Widget alignment (Default: Gtk.Align.START)
        expand : bool, optional
            Extra space will be divided evenly between all children that use
            this option (Default: False)
        fill : bool, optional
            Always allocated the full size of a Gtk.Box (Default: False)
        padding : int, optional
            Extra space in pixels to put between this child and its neighbors
            (Default: 0)

        Raises
        ------
        TypeError
            if align type is not Gtk.Align
        """

        if type(align) is not Gtk.Align:
            raise TypeError(
                "Wrong type for align, expected Gtk.Align")

        # Using default theme
        if not self.use_classic_theme:

            if align == Gtk.Align.END:
                self.headerbar.pack_end(widget)

            else:
                self.headerbar.pack_start(widget)

                if self.headerbar_image in self.headerbar.get_children():
                    self.headerbar.remove(self.headerbar_image)

        # Using classic theme
        else:

            if align == Gtk.Align.END:
                self.grid_tools.pack_end(widget, expand, fill, padding)
            else:
                self.grid_tools.pack_start(widget, expand, fill, padding)

    def add_button(self, label, response, align=Gtk.Align.START):
        """ Add a button to dialog interface

        Parameters
        ----------
        label : str
            Button label
        response : Gtk.ResponseType
            Button response type
        align : Gtk.Align, optional
            Button alignment (Default: Gtk.Align.START)

        Returns
        -------
        Gtk.Button
            Generated button to allow signals connecting when no parent
            available (Gtk.Window mode)

        Raises
        ------
        TypeError
            if align type is not Gtk.Align
        TypeError
            if response type is not Gtk.ResponseType
        ValueError
            if a button with response type already exists
        """

        if type(response) is not Gtk.ResponseType:
            raise TypeError(
                "Wrong type for response, expected Gtk.ResponseType")

        if type(align) is not Gtk.Align:
            raise TypeError(
                "Wrong type for align, expected Gtk.Align")

        if response in self.sensitive_data.keys():
            raise ValueError(
                "Response type %s already set" % str(response))

        # ------------------------------------
        #   Button
        # ------------------------------------

        button = Gtk.Button()
        button.set_label(str(label))

        self.sensitive_data[response] = button

        # ------------------------------------
        #   Manage themes
        # ------------------------------------

        # Add a style to button for specific responses
        if response in [Gtk.ResponseType.APPLY, Gtk.ResponseType.ACCEPT]:
            button.get_style_context().add_class(
                Gtk.STYLE_CLASS_SUGGESTED_ACTION)
        elif response in [Gtk.ResponseType.YES, Gtk.ResponseType.REJECT]:
            button.get_style_context().add_class(
                Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION)

        if align == Gtk.Align.END:
            self.grid_actions_buttons.pack_end(button, False, False, 0)
        else:
            self.grid_actions_buttons.pack_start(button, False, False, 0)

        # Gtk.Dialog
        if self.parent is not None:
            button.connect("clicked", self.emit_response, response)

        return button

    def add_help(self, data):
        """ Add a button to dialog interface

        Parameters
        ----------
        data : dict
            Help data dictionary
        """

        self.help_data = list()

        # Get help data from specified order
        for item in data["order"]:

            # Dictionary value
            if type(data[item]) is dict:
                self.help_data.append("<b>%s</b>" % item)

                for key, value in sorted(data[item].items(),
                                         key=lambda key: key[0]):

                    self.help_data.append("\t<b>%s</b>\n\t\t%s" % (
                        key.replace('>', "&gt;").replace('<', "&lt;"), value))

            # List value
            elif type(data[item]) is list:
                self.help_data.append("\n\n".join(data[item]))

            # String value
            elif type(data[item]) is str:
                self.help_data.append(data[item])

        # ------------------------------------
        #   Generate button data
        # ------------------------------------

        if not self.has_help_button:
            self.has_help_button = True

            self.button_help.connect("clicked", self.__on_show_help)

            self.grid_actions.pack_start(self.button_help, False, False, 0)

    def pack_end(self, child, expand=True, fill=True, padding=int()):
        """ Packing child widget into dialog grid

        Parameters
        ----------
        child : Gtk.Widget
            Child widget to pack
        expand : bool, optional
            Extra space will be divided evenly between all children that use
            this option (Default: True)
        fill : bool, optional
            Always allocated the full size of a Gtk.Box (Default: True)
        padding : int, optional
            Extra space in pixels to put between this child and its neighbors
            (Default: 0)
        """

        self.grid.pack_end(child, expand, fill, padding)

    def pack_start(self, child, expand=True, fill=True, padding=int()):
        """ Packing child widget into dialog grid

        Parameters
        ----------
        child : Gtk.Widget
            Child widget to pack
        expand : bool, optional
            Extra space will be divided evenly between all children that use
            this option (Default: True)
        fill : bool, optional
            Always allocated the full size of a Gtk.Box (Default: True)
        padding : int, optional
            Extra space in pixels to put between this child and its neighbors
            (Default: 0)
        """

        self.grid.pack_start(child, expand, fill, padding)

    def set_border_width(self, border_width):
        """ Set the dialog grid border width value

        Parameters
        ----------
        border_width : int
            Container border with in pixels
        """

        # Avoid to have buttons stick to main grid with a null border_width
        if border_width == 0:
            self.grid.set_spacing(0)
            self.grid_actions.set_border_width(12)

        self.grid.set_border_width(border_width)

    def set_spacing(self, spacing):
        """ Set the dialog grid spacing value

        Parameters
        ----------
        spacing : int
            Number of pixels between each child
        """

        self.grid.set_spacing(spacing)

    def get_size(self):
        """ Get current dialog size

        Returns
        -------
        tuple
            dialog size
        """

        return self.window.get_size()

    def set_size(self, width, height):
        """ Set default size for dialog

        Parameters
        ----------
        width : int
            Dialog width
        height : int
            Dialog height
        """

        self.window.set_default_size(width, height)

    def set_size_request(self, width, height):
        """ Set a new size for dialog

        Parameters
        ----------
        width : int
            Dialog width
        height : int
            Dialog height
        """

        self.window.set_size_request(width, height)

    def set_modal(self, modal):
        """ Set dialog modal status

        Parameters
        ----------
        modal : bool
            New dialog modal status
        """

        self.window.set_modal(modal)

    def set_orientation(self, orientation):
        """ Set dialog grid orientation

        Parameters
        ----------
        orientation : Gtk.Orientation
            Dialog grid orientation

        Raises
        ------
        TypeError
            if orientation type is not Gtk.Orientation
        """

        if type(orientation) is not Gtk.Orientation:
            raise TypeError(
                "Wrong type for orientation, expected Gtk.Orientation")

        self.grid.set_orientation(orientation)

    def set_resizable(self, resizable):
        """ Set dialog resizable mode

        Parameters
        ----------
        resizable : bool
            Dialog resizable status
        """

        self.window.set_resizable(resizable)

    def set_default_response(self, response):
        """ Set default button response

        Parameters
        ----------
        response : Gtk.ResponseType
            Button response type
        """

        if response not in self.sensitive_data.keys():
            raise NameError(
                "%s type did not exists in data dictionary" % str(response))

        self.window.set_default_response(response)

    def set_response_sensitive(self, response, sensitive):
        """ Set button sensitive status

        Parameters
        ----------
        response : Gtk.ResponseType
            Button response type
        sensitive : bool
            Button sensitive status

        Raises
        ------
        NameError
            if button response type has not been set previously
        """

        if response not in self.sensitive_data.keys():
            raise NameError(
                "%s type did not exists in data dictionary" % str(response))

        self.sensitive_data[response].set_sensitive(sensitive)

    def set_subtitle(self, subtitle):
        """ Set headerbar subtitle if available

        Parameters
        ----------
        subtitle : str
            Headerbar subtitle string
        """

        if not self.use_classic_theme:
            self.headerbar.set_subtitle(subtitle)

    def emit_response(self, widget, response):
        """ Close dialog and emit specified response

        Parameters
        ----------
        widget : Gtk.Button
            Object which received the signal
        response : Gtk.ResponseType
            Response to emit when pushing button

        Raises
        ------
        TypeError
            if response type is not Gtk.ResponseType
        """

        if type(response) is not Gtk.ResponseType:
            raise TypeError(
                "Wrong type for response, expected Gtk.ResponseType")

        self.window.response(response)

    def on_activate_listboxrow(self, widget, row):
        """ Activate internal widget when a row has been activated

        Parameters
        ----------
        widget : Gtk.ListBox
            Object which receive signal
        row : Gtk.ListBoxRow
            Activated row
        """

        data = {
            Gtk.ComboBox: "popup",
            Gtk.Entry: "grab_focus",
            Gtk.FileChooserButton: "activate",
            Gtk.FontButton: "clicked",
            Gtk.MenuButton: "clicked",
            Gtk.SpinButton: "grab_focus",
        }

        widget = row.get_widget()
        widget_type = type(widget)

        if widget_type in data.keys():
            self.__on_execute_method_from_widget(widget, data.get(widget_type))

        elif widget_type is Gtk.Switch:
            widget.set_active(not widget.get_active())

        elif widget_type is Gtk.Box:
            children = widget.get_children()

            if children:
                self.__on_execute_method_from_widget(
                    children[0], data.get(type(children[0])))


class HelpDialog(CommonWindow):

    def __init__(self, parent, title, message, icon):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        title : str
            Dialog title
        message : str
            Dialog message
        icon : str
            Default icon name
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self, parent, title, icon, classic_theme)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.message = message

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, 480)

        self.set_resizable(True)

        # ------------------------------------
        #   Scrollview
        # ------------------------------------

        scroll = Gtk.ScrolledWindow()
        view = Gtk.Viewport()

        # ------------------------------------
        #   Message
        # ------------------------------------

        text = Gtk.Label()

        # Properties
        text.set_line_wrap(True)
        text.set_use_markup(True)
        text.set_max_width_chars(10)
        text.set_markup(self.message)
        text.set_halign(Gtk.Align.FILL)
        text.set_valign(Gtk.Align.START)
        text.set_justify(Gtk.Justification.FILL)
        text.set_line_wrap_mode(Pango.WrapMode.WORD)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        view.add(text)
        scroll.add(view)

        self.pack_start(scroll, True, True)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.show_all()



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.engine.api import GEM
from geode_gem.engine.utils import get_data

from geode_gem.ui.data import Metadata

# GObject
from gi.repository import Gdk, GLib, Gtk, Pango

# System
from time import sleep

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Message(Gtk.MessageDialog):

    def __init__(self, title, text):
        """ Constructor

        Parameters
        ----------
        title : str
            Main text of message dialog
        text : str
            Secondary text of message dialog
        """

        super().__init__()

        self.set_transient_for(None)

        self.set_markup(title)
        self.format_secondary_text(text)

        self.add_button(_("Close"), Gtk.ResponseType.CLOSE)

    def run(self):
        """ Launch and show the message dialog
        """

        super().run()

        self.destroy()


class Splash(Gtk.Window):

    def __init__(self, api):
        """ Constructor

        Parameters
        ----------
        api : gem.engine.api.GEM
            GEM API instance

        Raises
        ------
        TypeError
            if api type is not gem.engine.api.GEM
            if metadata type is not gem.engine.lib.configuration.Configuration
        """

        if not type(api) is GEM:
            raise TypeError("Wrong type for api, expected gem.engine.api.GEM")

        Gtk.Window.__init__(self)

        # ------------------------------------
        #   Initialize variables
        # ------------------------------------

        self.main_loop = None

        # ------------------------------------
        #   Initialize API
        # ------------------------------------

        # GEM API
        self.api = api

        # Quick access to API logger
        self.logger = api.logger

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Init packing
        self.__init_packing()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Load widgets into main interface
        """

        # ------------------------------------
        #   Main window
        # ------------------------------------

        self.set_title(Metadata.NAME)

        self.set_modal(True)
        self.set_can_focus(True)
        self.set_resizable(False)
        self.set_keep_above(True)
        self.set_skip_taskbar_hint(True)
        self.set_type_hint(Gdk.WindowTypeHint.SPLASHSCREEN)

        self.set_default_size(380, -1)
        self.set_position(Gtk.WindowPosition.CENTER)

        # ------------------------------------
        #   Grid
        # ------------------------------------

        self.grid = Gtk.Box()

        # Properties
        self.grid.set_spacing(4)
        self.grid.set_border_width(16)
        self.grid.set_homogeneous(False)
        self.grid.set_orientation(Gtk.Orientation.VERTICAL)

        # ------------------------------------
        #   Logo
        # ------------------------------------

        self.label_splash = Gtk.Label()
        self.image_splash = Gtk.Image()

        # Properties
        self.label_splash.set_line_wrap(True)
        self.label_splash.set_use_markup(True)
        self.label_splash.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.label_splash.set_markup(
            "<span weight='bold' size='x-large'>%s - %s</span>\n<i>%s</i>" % (
                Metadata.NAME,
                Metadata.VERSION,
                Metadata.CODE_NAME))

        self.image_splash.set_from_file(
            str(get_data("data", "desktop", "gem.svg")))
        self.image_splash.set_pixel_size(256)

        # ------------------------------------
        #   Progressbar
        # ------------------------------------

        self.label_progress = Gtk.Label()

        self.progressbar = Gtk.ProgressBar()

        # Properties
        self.label_progress.set_text(_("Migrating entries from old database"))
        self.label_progress.set_line_wrap_mode(Pango.WrapMode.WORD)
        self.label_progress.set_line_wrap(True)
        self.label_progress.set_no_show_all(True)

        self.progressbar.set_show_text(True)
        self.progressbar.set_no_show_all(True)

    def __init_packing(self):
        """ Initialize widgets packing in main window
        """

        self.grid.pack_start(self.image_splash, True, True, 0)
        self.grid.pack_start(self.label_splash, False, False, 8)
        self.grid.pack_start(self.label_progress, False, False, 8)
        self.grid.pack_start(self.progressbar, False, False, 0)

        self.add(self.grid)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.set_auto_startup_notification(False)

        self.show_all()

        self.refresh()

        # ------------------------------------
        #   Check migration
        # ------------------------------------

        GLib.idle_add(self.check_database_migration().__next__)

        # ------------------------------------
        #   Main loop
        # ------------------------------------

        self.main_loop = GLib.MainLoop()
        self.main_loop.run()

    def check_database_migration(self):
        """ Call API check_database method
        """

        # Trick to show splash image
        sleep(0.001)

        yield True
        for data in self.api.check_database():
            if data is not None:
                self.update(*data)
                yield True

        GLib.idle_add(self.close)
        yield False

    def close(self):
        """ Stop interface
        """

        self.set_auto_startup_notification(True)

        # Sleep to avoid ultra quick splash
        sleep(0.42)

        # Rare case where the mainloop is not init when close() is running
        if self.main_loop is not None:
            self.main_loop.quit()

        self.destroy()

    def update(self, index, length):
        """ Update progress in progressbar widgets

        Parameters
        ----------
        index : int
            Current progession index step
        length : int
            Progression max iterations number
        """

        self.refresh()

        if not index == 0 and index <= length:
            self.label_progress.show()
            self.progressbar.show()
            self.progressbar.set_text("%d / %d" % (index, length))
            self.progressbar.set_fraction(float(index) / length)

            self.refresh()

    def refresh(self):
        """ Refresh all pendings event in main interface
        """

        while Gtk.events_pending():
            Gtk.main_iteration()



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GEM
from geode_gem.ui.data import Icons, Folders
from geode_gem.ui.widgets.window import CommonWindow

# GObject
from gi.repository import Gtk, Pango

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CleanCacheDialog(CommonWindow):

    def __init__(self, parent):
        """ Constructor

        Parameters
        ----------
        parent : Gtk.Window
            Parent object
        """

        classic_theme = False
        if parent is not None:
            classic_theme = parent.use_classic_theme

        CommonWindow.__init__(self,
                              parent,
                              _("Clean icons cache"),
                              Icons.Symbolic.FOLDER,
                              classic_theme)

        # ------------------------------------
        #   Prepare interface
        # ------------------------------------

        # Init widgets
        self.__init_widgets()

        # Start interface
        self.__start_interface()

    def __init_widgets(self):
        """ Initialize interface widgets
        """

        self.set_size(640, -1)

        self.set_spacing(6)

        self.set_resizable(True)

        # ------------------------------------
        #   Title
        # ------------------------------------

        self.label_title = Gtk.Label()

        # Properties
        self.label_title.set_markup(
            "<span weight='bold' size='large'>%s</span>" % (
                _("Do you really want to clean icons cache directory ?")))
        self.label_title.set_use_markup(True)
        self.label_title.set_halign(Gtk.Align.CENTER)
        self.label_title.set_ellipsize(Pango.EllipsizeMode.END)

        # ------------------------------------
        #   Description
        # ------------------------------------

        self.label_description = Gtk.Label()

        # Properties
        self.label_description.set_text(str(Folders.CACHE))
        self.label_description.set_line_wrap(True)
        self.label_description.set_max_width_chars(8)
        self.label_description.set_single_line_mode(False)
        self.label_description.set_justify(Gtk.Justification.FILL)
        self.label_description.set_line_wrap_mode(Pango.WrapMode.WORD)

        # ------------------------------------
        #   Integrate widgets
        # ------------------------------------

        self.pack_start(self.label_title, False, False)
        self.pack_start(self.label_description, False, False)

    def __start_interface(self):
        """ Load data and start interface
        """

        self.add_button(_("No"), Gtk.ResponseType.NO)
        self.add_button(_("Yes"), Gtk.ResponseType.YES, Gtk.Align.END)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.ui.data import Icons
from geode_gem.widgets.common import GeodeGtkCommon
from geode_gem.widgets import GeodeGtk

# GObject
from gi.repository import Gtk

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGEMPlaceholder(GeodeGtkCommon, Gtk.ScrolledWindow):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        GeodeGtkCommon.__init__(self, Gtk.ScrolledWindow, **kwargs)

        self.inner_grid = GeodeGtk.Box(
            GeodeGtk.Image(
                is_expandable=True,
                is_fillable=True,
                set_from_icon_name=(Icons.Symbolic.GAMING,
                                    Gtk.IconSize.DIALOG),
                set_halign=Gtk.Align.CENTER,
                set_pixel_size=256,
                set_valign=Gtk.Align.END,
                set_style=Gtk.STYLE_CLASS_DIM_LABEL,
            ),
            GeodeGtk.Label(
                is_expandable=True,
                is_fillable=True,
                set_halign=Gtk.Align.CENTER,
                set_valign=Gtk.Align.START,
                set_text=_("Start to play by drag & drop some files into "
                           "interface"),
            ),
            set_border_width=18,
            set_orientation=Gtk.Orientation.VERTICAL,
            set_spacing=12,
        )

        # Packing
        self.add(self.inner_grid)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# GObject
from gi import require_version
require_version("Gtk", "3.0")

from gi.repository import Gtk

# Logging
from logging import getLogger


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class GeodeGtkCommon():

    __setters__ = None

    def __init__(self, subclass, **kwargs):
        """ Constructor
        """

        if subclass is None:
            raise ValueError("Cannot use None as subclass")

        subclass.__init__(self)

        self.inner_grid = None
        self.inner_widgets = dict()

        # Uniq identifier
        self.identifier = kwargs.get("identifier", str(id(self)))
        # Widget class style
        self.current_style = None

        arguments = self.__setters__.copy() if self.__setters__ else dict()
        arguments.update(kwargs)

        for key, value in arguments.items():
            self.__set_widget(key, value)

        self.is_expandable = kwargs.get("is_expandable", False)
        self.is_fillable = kwargs.get("is_fillable", False)

    def __set_widget(self, method, value):
        """ Use Gtk widget methods if this one exists

        Parameters
        ----------
        method : str
            Setter method name
        value : object
            Object to use as setter method parameters

        Returns
        -------
        object
            Method result, None otherwise
        """

        if not method.startswith("set_"):
            return None

        if not hasattr(self, method):
            getLogger("gem").warning(
                f"Cannot found '{method}' in '{self.identifier}' widget")
            return None

        if value is None:
            return None

        if type(value) not in (list, tuple):
            value = (value, )

        return getattr(self, method)(*value)

    def do_show(self):
        """ Virtual method called when self.show() method is called
        """

        if hasattr(self, "do_show"):
            super().do_show(self)

        # Ensure to show all internal widgets
        if isinstance(self.inner_grid, Gtk.Container):
            self.inner_grid.show_all()

    def append_widget(self, widget, identifier=None):
        """ Check if a specific widget exists in internal container

        Parameters
        ----------
        widget : Gtk.Widget
            Gtk object instance
        identifier : str, optional
            Set a specific identifier to specified widget
        """

        if identifier is not None:
            setattr(widget, "identifier", identifier)

        if hasattr(widget, "identifier") and widget.identifier is not None:
            self.inner_widgets[widget.identifier] = widget

        if hasattr(widget, "inner_widgets"):
            self.inner_widgets.update(widget.inner_widgets)

    def has_widget(self, widget_key):
        """ Check if a specific widget exists in internal container

        Parameters
        ----------
        widget_key : str
            Internal widget keys, contains in self.inner_widgets

        Returns
        -------
        bool
            True if widget exists, False otherwise
        """

        return widget_key in self.inner_widgets.keys()

    def get_widget(self, widget_key):
        """ Retrieve a specific widget from internal container

        Parameters
        ----------
        widget_key : str
            Internal widget keys, contains in self.inner_widgets

        Returns
        -------
        Gtk.Widget or None
            Widget from internal container

        Raises
        ------
        KeyError
            When specified widget do not exists in widget
        """

        if not self.has_widget(widget_key):
            raise KeyError(f"Cannot found {widget_key} in {self} widgets")

        return self.inner_widgets.get(widget_key, None)

    def get_active(self, widget=None):
        """ Returns the widget active entry

        Parameters
        ----------
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets

        Returns
        -------
        Gtk.Widget
            The activated widget from internal container
        """

        if widget is None:
            return super().get_active()

        elif self.has_widget(widget):
            return self.get_widget(widget).get_active()

        return None

    def set_active(self, value, widget=None):
        """ Set the active value for a specific widget

        Parameters
        ----------
        value : bool or int
            The new activate value for specified widget
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets
        """

        if widget is None:
            super().set_active(value)

        elif self.has_widget(widget):
            self.get_widget(widget).set_active(value)

    def get_sensitive(self, widget=None):
        """ Returns the widget sensitivity

        Parameters
        ----------
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets

        Returns
        -------
        bool
            True if the widget is sensitive, False otherwise
        """

        if widget is None:
            return super().get_sensitive()

        elif self.has_widget(widget):
            return self.get_widget(widget).get_sensitive()

        return True

    def set_sensitive(self, sensitive, widget=None):
        """ Set the widget sensitivity

        Parameters
        ----------
        sensitive : bool
            The new internal widget sensitive status
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets
        """

        if widget is None:
            super().set_sensitive(sensitive)

        elif self.has_widget(widget):
            self.get_widget(widget).set_sensitive(sensitive)

    def get_visible(self, widget=None):
        """ Returns the widget sensitivity

        Parameters
        ----------
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets

        Returns
        -------
        bool
            True if the widget is visible, False otherwise
        """

        if widget is None:
            return super().get_visible()

        elif self.has_widget(widget):
            return self.get_widget(widget).get_visible()

        return True

    def set_visible(self, visible, widget=None):
        """ Set the widget sensitivity

        Parameters
        ----------
        visible : bool
            The new internal widget visible status
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets
        """

        if widget is None:
            super().set_visible(visible)

        elif self.has_widget(widget):
            self.get_widget(widget).set_visible(visible)

    def has_style(self, style, widget=None):
        """ Check if a specific class is available in specified widget

        Parameters
        ----------
        style : str
            Class style name
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets

        Returns
        -------
        bool
            True if the widget has class, False otherwise
        """

        return style in self.get_style(widget=widget)

    def get_style(self, widget=None):
        """ Retrieve classes list for specified widget

        Parameters
        ----------
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets

        Returns
        -------
        list
            Widget classes string list
        """

        if widget is not None and self.has_widget(widget):
            context = self.get_widget(widget).get_style_context()
        else:
            context = self.get_style_context()

        return context.list_classes()

    def set_style(self, style=None, widget=None):
        """ Define a specific class for specified widget

        Parameters
        ----------
        style : str, optional
            Class style name
        widget : str, optionnal
            Internal widget keys, contains in self.inner_widgets
        """

        if widget is not None and self.has_widget(widget):
            context = self.get_widget(widget).get_style_context()
        else:
            context = self.get_style_context()

        if self.current_style is not None and not self.current_style == style:
            context.remove_class(self.current_style)

        if style is not None and not self.current_style == style:
            context.add_class(style)

        self.current_style = style



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gtk


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonComboBox(GeodeGtkCommon):

    def __init__(self, subclass, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        subclass : Gtk.Entry
            Subclass widget type
        """

        GeodeGtkCommon.__init__(self, subclass, **kwargs)

        # Packing
        for element in args:
            self.append_widget(element)

    def append_widget(self, widget, *args, **kwargs):
        """ See geode_gem.widgets.common.GeodeGtkCommon.append_widget
        """

        super().append_widget(widget, *args, **kwargs)

        if isinstance(widget, Gtk.TreeModel):
            self.set_model(widget)

        elif isinstance(widget, Gtk.CellRenderer):
            self.pack_start(widget, widget.is_expandable)

            for attribute, value in widget.attributes.items():
                self.add_attribute(widget, attribute, value)


class GeodeGtkComboBox(CommonComboBox, Gtk.ComboBox):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonComboBox.__init__(self, Gtk.ComboBox, *args, **kwargs)


class GeodeGtkComboBoxText(CommonComboBox, Gtk.ComboBoxText):

    def __init__(self, *args, **kwargs):
        """ Constructor
        """

        CommonComboBox.__init__(self, Gtk.ComboBoxText, *args, **kwargs)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Datetime
from datetime import timedelta

# Geode
from geode_gem.ui.data import Columns
from geode_gem.ui.utils import string_from_date, string_from_time
from geode_gem.ui.placeholder import GeodeGEMPlaceholder
from geode_gem.widgets import GeodeGtk
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gdk, GdkPixbuf, Gtk, Pango

# Python
from enum import Enum

# Translation
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonGamesView(GeodeGtkCommon, Gtk.ScrolledWindow):

    def __init__(self, interface, view, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        interface : geode_gem.ui.interface.MainWindow
            Main interface instance
        view : geode_gem.ui.view.CommonGamesView
            View instance
        """

        GeodeGtkCommon.__init__(self, Gtk.ScrolledWindow, **kwargs)

        self.interface = interface
        self.view = view

        self.storage = dict()

        # ------------------------------------
        #   Packing
        # ------------------------------------

        self.add(self.view)
        self.append_widget(self.view)

    def clear(self):
        """ Remove all items from games view
        """

        self.view.unselect_all()
        self.view.clear()

        self.storage.clear()

    def append_item(self, identifier, data):
        """ Append a new item in games view

        Parameters
        ----------
        identifier : str
            Treeiter identifier
        data : list
            Data structure for specific view
        """

        self.storage[identifier] = self.view.append(data)

        return self.storage.get(identifier)

    def has_item(self, identifier):
        """ Check if a specific identifier exists in main storage

        Parameters
        ----------
        identifier : str
            Item identifier used in main storage

        Returns
        -------
        bool
            True if item exists in main storage, False otherwise
        """

        return identifier in self.storage

    def remove_item(self, identifier):
        """ Remove an item from games view

        Parameters
        ----------
        identifier : str
            Treeiter identifier
        """

        if identifier in self.storage:
            self.view.remove(self.storage.get(identifier))

            del self.storage[identifier]

    def get_treeiter(self, identifier):
        """ Retrieve an item from main storage for a specific identifier

        Parameters
        ----------
        identifier : str
            Item identifier used in main storage

        Returns
        -------
        Gtk.TreeIter
            Item instance is found, None otherwise
        """

        return self.storage.get(identifier, None)


class GeodeGEMViews(GeodeGtkCommon, Gtk.Box):

    __target__ = [
        Gtk.TargetEntry.new("text/uri-list", 0, 1337)
    ]

    class Name(Enum):

        LIST = "list"
        GRID = "grid"

    def __init__(self, interface, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        interface : geode_gem.ui.interface.MainWindow
            Main interface instance
        """

        GeodeGtkCommon.__init__(self, Gtk.Box, **kwargs)

        self.interface = interface
        self.logger = interface.logger

        self.games_identifier_storage = list()

        self.visible_view = None

        # ------------------------------------
        #   Properties
        # ------------------------------------

        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.set_spacing(0)

        # ------------------------------------
        #   Widgets
        # ------------------------------------

        self.view_placeholder = GeodeGEMPlaceholder(self.interface)
        self.view_list = GeodeGEMTreeView(self.interface)
        self.view_grid = GeodeGEMIconView(self.interface)

        # ------------------------------------
        #   Drag and drop
        # ------------------------------------

        self.drag_dest_set(Gtk.DestDefaults.MOTION | Gtk.DestDefaults.DROP,
                           self.__target__,
                           Gdk.DragAction.COPY)

        self.view_list.view.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            self.__target__,
            Gdk.DragAction.COPY)

        self.view_grid.view.drag_source_set(
            Gdk.ModifierType.BUTTON1_MASK,
            self.__target__,
            Gdk.DragAction.COPY)

        # ------------------------------------
        #   Signals
        # ------------------------------------

        self.connect(
            "drag-data-received", interface.on_drag_data_to_main_interface)

        # ------------------------------------
        #   Packing
        # ------------------------------------

        self.pack_start(self.view_placeholder, True, True, 0)

        self.pack_start(self.view_list, True, True, 0)
        self.append_widget(self.view_list)

        self.pack_start(self.view_grid, True, True, 0)
        self.append_widget(self.view_grid)

    @property
    def iconview(self):
        """ Retrieve iconview instance

        Returns
        -------
        GeodeGtk.IconView
            Iconview instance
        """

        return self.view_grid.view

    @property
    def placeholder(self):
        """ Retrieve placeholder instance

        Returns
        -------
        GeodeGEM.Placeholder
            Placeholder instance
        """

        return self.view_placeholder

    @property
    def treeview(self):
        """ Retrieve treeview instance

        Returns
        -------
        GeodeGtk.TreeView
            Treeview instance
        """

        return self.view_list.view

    def clear(self):
        """ Clear views and lists storages
        """

        self.games_identifier_storage.clear()

        self.logger.debug("Clear games treeview content")
        self.view_list.clear()

        self.logger.debug("Clear games iconview content")
        self.view_grid.clear()

    def append_games(self, console, games):
        """ Append games in both views

        Parameters
        ----------
        console : geode_gem.engine.console.Console
            Console instance
        games : list
            Games object list
        """

        self.treeview.freeze_child_notify()

        for index, game in enumerate(games, start=1):
            self.append_game(console, game)

            self.treeview.thaw_child_notify()
            yield index, game
            self.treeview.freeze_child_notify()

        self.treeview.thaw_child_notify()

    def append_game(self, console, game):
        """ Append a game to games views

        Parameters
        ----------
        console : geode_gem.engine.console.Console
            Console instance
        game : geode_gem.engine.game.Game
            Game instance
        """

        self.view_grid.append_item(console, game)
        self.view_list.append_item(console, game)

        self.games_identifier_storage.append(game.id)

    def get_selected_game(self):
        """ Retrieve selected game from visible view

        Returns
        -------
        Gtk.TreeIter
            Selected game if found, None otherwise
        """

        treeiter = None

        if self.visible_view == GeodeGEMViews.Name.LIST:
            model = self.treeview.get_model()
            treeiter = self.treeview.get_selected_treeiter()

            if treeiter is not None:
                return model.get_value(treeiter, Columns.List.OBJECT)

        elif self.visible_view == GeodeGEMViews.Name.GRID:
            model = self.iconview.get_model()
            treeiter = self.iconview.get_selected_treeiter()

            if treeiter is not None:
                return model.get_value(treeiter, Columns.Grid.OBJECT)

        return None

    def get_iter_from_key(self, identifier):
        """ Retrieve views iters for a specific game identifier

        Parameters
        ----------
        identifier : str
            Game identifier
        """

        return [getattr(self, f"view_{key}").get_treeiter(identifier)
                for key in ("list", "grid")]

    def has_game(self, identifier):
        """ Check if a specific identifier exists in main storage

        Parameters
        ----------
        identifier : str
            Game identifier used in main storage

        Returns
        -------
        bool
            True if game exists in main storage, False otherwise
        """

        return identifier in self.games_identifier_storage

    def refilter(self):
        """ Refilter games views
        """

        self.treeview.refilter()
        self.iconview.refilter()

    def remove_game(self, identifier):
        """ Remove a game from games views

        Parameters
        ----------
        identifier : str
            Game identifier
        """

        if identifier in self.games_identifier_storage:
            self.games_identifier_storage.remove(identifier)

        self.view_list.remove_item(identifier)
        self.view_grid.remove_item(identifier)

    def set_placeholder_visibility(self, is_visible):
        """ Define placeholder visibility

        Parameters
        ----------
        is_visible : bool
            Placeholder visibility status
        """

        if is_visible:
            self.view_list.hide()
            self.view_grid.hide()
            self.view_placeholder.show_all()

        else:
            self.set_view(self.visible_view)
            self.view_placeholder.hide()

        self.view_placeholder.set_visible(is_visible)

    def set_view(self, identifier, show=True):
        """ Set the games view to show

        Parameters
        ----------
        identifier : str
            View identifier
        show : boolean, optional
            Show the specified view when set as visible_view
        """

        if not isinstance(identifier, GeodeGEMViews.Name):
            raise KeyError(
                f"Cannot found '{identifier}' in GeodeGEMViews.Name")

        for key in GeodeGEMViews.Name:
            widget = getattr(self, f"view_{key.value}")

            widget.set_visible(key is identifier and show)
            if widget.get_visible():
                widget.show_all()

        self.visible_view = identifier

    def unselect_all(self):
        """ Unselect selections from both games views
        """

        self.logger.debug("Unselect games on both views")
        self.iconview.unselect_all()
        self.treeview.unselect_all()


class GeodeGEMTreeView(CommonGamesView):

    def __init__(self, interface, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        interface : geode_gem.ui.interface.MainWindow
            Main interface instance
        """

        CommonGamesView.__init__(
            self,
            interface,
            GeodeGtk.TreeView(
                Gtk.ListStore(
                    str,                # Favorite icon
                    str,                # Multiplayer icon
                    str,                # Finish icon
                    str,                # Name
                    int,                # Played
                    str,                # Last play
                    str,                # Last time play
                    str,                # Time play
                    int,                # Score
                    str,                # Installed
                    str,                # Custom parameters
                    str,                # Screenshots
                    str,                # Save states
                    object,             # Game object
                    GdkPixbuf.Pixbuf    # Thumbnail
                ),
                GeodeGtk.TreeViewColumn(
                    None,
                    GeodeGtk.CellRendererPixbuf(
                        attributes={
                            "icon-name": Columns.List.FAVORITE,
                        },
                        set_padding=(4, 0),
                    ),
                    identifier="favorite",
                    set_resizable=False,
                    set_sizing=Gtk.TreeViewColumnSizing.FIXED,
                    sort_column_id=Columns.List.FAVORITE,
                ),
                GeodeGtk.TreeViewColumn(
                    None,
                    GeodeGtk.CellRendererPixbuf(
                        attributes={
                            "icon-name": Columns.List.MULTIPLAYER,
                        },
                        set_padding=(4, 0),
                    ),
                    identifier="multiplayer",
                    set_resizable=False,
                    set_sizing=Gtk.TreeViewColumnSizing.FIXED,
                    sort_column_id=Columns.List.MULTIPLAYER,
                ),
                GeodeGtk.TreeViewColumn(
                    None,
                    GeodeGtk.CellRendererPixbuf(
                        attributes={
                            "icon-name": Columns.List.FINISH,
                        },
                        set_padding=(4, 0),
                    ),
                    identifier="finish",
                    set_resizable=False,
                    set_sizing=Gtk.TreeViewColumnSizing.FIXED,
                    sort_column_id=Columns.List.FINISH,
                ),
                GeodeGtk.TreeViewColumn(
                    _("Name"),
                    GeodeGtk.CellRendererPixbuf(
                        attributes={
                            "pixbuf": Columns.List.THUMBNAIL,
                        },
                        set_padding=(2, 0),
                    ),
                    GeodeGtk.CellRendererText(
                        attributes={
                            "text": Columns.List.NAME,
                        },
                        properties={
                            "ellipsize": Pango.EllipsizeMode.END
                        },
                        is_expandable=True,
                        set_alignment=(0, 0.5),
                        set_padding=(4, 4),
                    ),
                    identifier="name",
                    set_alignment=0,
                    set_expand=True,
                    set_min_width=100,
                    set_fixed_width=300,
                    sort_column_id=Columns.List.NAME,
                ),
                GeodeGtk.TreeViewColumn(
                    _("Launch"),
                    GeodeGtk.CellRendererText(
                        attributes={
                            "text": Columns.List.PLAYED,
                        },
                        set_padding=(4, 4),
                    ),
                    identifier="play",
                    sort_column_id=Columns.List.PLAYED,
                ),
                GeodeGtk.TreeViewColumn(
                    _("Last launch"),
                    GeodeGtk.CellRendererText(
                        attributes={
                            "text": Columns.List.LAST_PLAY,
                        },
                        set_alignment=(0, .5),
                        set_padding=(4, 0),
                    ),
                    GeodeGtk.CellRendererText(
                        attributes={
                            "text": Columns.List.LAST_TIME_PLAY,
                        },
                        set_alignment=(1, .5),
                        set_padding=(4, 0),
                    ),
                    identifier="last_play",
                    sort_column_id=Columns.List.LAST_PLAY,
                ),
                GeodeGtk.TreeViewColumn(
                    _("Play time"),
                    GeodeGtk.CellRendererText(
                        attributes={
                            "text": Columns.List.TIME_PLAY,
                        },
                        set_padding=(4, 0),
                    ),
                    identifier="play_time",
                    sort_column_id=Columns.List.TIME_PLAY,
                ),
                GeodeGtk.TreeViewColumn(
                    _("Score"),
                    *[
                        GeodeGtk.CellRendererPixbuf(
                            identifier=f"cell_score_{index}",
                            is_expandable=True,
                            set_padding=(2, 0),
                        ) for index in range(1, 6)
                    ],
                    identifier="score",
                    cell_data_func=self.on_update_cells,
                    sort_column_id=Columns.List.SCORE,
                ),
                GeodeGtk.TreeViewColumn(
                    _("Installed"),
                    GeodeGtk.CellRendererText(
                        attributes={
                            "text": Columns.List.INSTALLED,
                        },
                        set_padding=(4, 0),
                    ),
                    identifier="installed",
                    sort_column_id=Columns.List.INSTALLED,
                ),
                GeodeGtk.TreeViewColumn(
                    _("Flags"),
                    GeodeGtk.CellRendererPixbuf(
                        attributes={
                            "icon-name": Columns.List.PARAMETER,
                        },
                        is_expandable=True,
                        set_padding=(2, 0),
                    ),
                    GeodeGtk.CellRendererPixbuf(
                        attributes={
                            "icon-name": Columns.List.SCREENSHOT,
                        },
                        is_expandable=True,
                        set_padding=(2, 0),
                    ),
                    GeodeGtk.CellRendererPixbuf(
                        attributes={
                            "icon-name": Columns.List.SAVESTATE,
                        },
                        is_expandable=True,
                        set_padding=(2, 0),
                    ),
                    identifier="flags",
                ),
                identifier="games",
                filterable=True,
                sorterable=True,
                set_enable_search=False,
                set_has_tooltip=True,
                set_search_column=Columns.List.NAME,
                set_show_expanders=False,
                sort_func=interface.on_sort_games_treeview,
                visible_func=interface.check_game_is_visible,
            ),
            identifier="treeview",
            **kwargs
        )

    def append_item(self, console, game):
        """ See geode_gem.ui.view.CommonGamesView.append_item

        Parameters
        ----------
        console : geode_gem.engine.console.Console
            Console instance
        game : geode_gem.engine.game.Game
            Game instance
        """

        thumbnail = self.interface.get_pixbuf_from_cache(
            "games", 22, game.id, game.cover)
        if not thumbnail or (game.cover and not game.cover.exists()):
            thumbnail = self.interface.icons_games_views.get("treeview")

        row_data = [
            self.interface.get_ui_icon(
                Columns.List.FAVORITE, game.favorite),
            self.interface.get_ui_icon(
                Columns.List.MULTIPLAYER, game.multiplayer),
            self.interface.get_ui_icon(
                Columns.List.FINISH, game.finish),
            game.name,
            game.played,
            (string_from_date(game.last_launch_date)
                if not game.last_launch_date.strftime("%d%m%y") == "010101"
                else None),
            (string_from_time(game.last_launch_time)
                if not game.last_launch_time == timedelta() else None),
            (string_from_time(game.play_time)
                if not game.play_time == timedelta() else None),
            game.score,
            string_from_date(game.installed) if game.installed else None,
            self.interface.get_ui_icon(
                Columns.List.PARAMETER,
                not game.emulator == console.emulator or game.default),
            self.interface.get_ui_icon(
                Columns.List.SCREENSHOT, game.screenshots),
            self.interface.get_ui_icon(
                Columns.List.SAVESTATE, game.savestates),
            game,
            thumbnail,
        ]

        return super().append_item(game.id, row_data)

    def on_update_cells(self, column, cell, model, treeiter, *args):
        """ Manage specific columns behavior during games adding

        Parameters
        ----------
        column : Gtk.TreeViewColumn
            Treeview column which contains cell
        cell : Gtk.CellRenderer
            Cell that is being rendered by column
        model : Gtk.TreeModel
            Rendered model
        treeiter : Gtk.TreeIter
            Rendered row
        """

        if not column.get_visible():
            return

        if cell.identifier.startswith("cell_score"):
            score = model.get_value(treeiter, Columns.List.SCORE)

            for index in range(1, 6):
                widget = self.view.get_widget(f"cell_score_{index}")

                icon_name = self.interface.get_ui_icon(
                    Columns.List.SCORE, score >= index)

                widget.set_sensitive(score >= index)
                widget.set_property("icon-name", icon_name)


class GeodeGEMIconView(CommonGamesView):

    def __init__(self, interface, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        interface : geode_gem.ui.interface.MainWindow
            Main interface instance
        """

        CommonGamesView.__init__(
            self,
            interface,
            GeodeGtk.IconView(
                Gtk.ListStore(
                    GdkPixbuf.Pixbuf,   # Cover icon
                    str,                # Name
                    object              # Game object
                ),
                identifier="games",
                filterable=True,
                sorterable=True,
                set_pixbuf_column=0,
                set_text_column=1,
                set_item_width=96,
                set_spacing=6,
                set_has_tooltip=True,
                sort_func=interface.on_sort_games_iconview,
                visible_func=interface.check_game_is_visible,
            ),
            identifier="iconview",
            **kwargs
        )

    def append_item(self, console, game):
        """ See geode_gem.ui.view.CommonGamesView.append_item

        Parameters
        ----------
        console : geode_gem.engine.console.Console
            Console instance
        game : geode_gem.engine.game.Game
            Game instance
        """

        thumbnail = self.interface.get_pixbuf_from_cache(
            "games", 96, game.id, game.cover)
        if not thumbnail or (game.cover and not game.cover.exists()):
            thumbnail = self.interface.icons_games_views.get("iconview")

        return super().append_item(game.id, [thumbnail, game.name, game])



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Datetime
from datetime import date, datetime, timedelta

# Filesystem
from pathlib import Path

# GEM
from geode_gem.engine.utils import (get_creation_datetime,
                                    generate_identifier,
                                    parse_timedelta)
from geode_gem.engine.emulator import Emulator

# Regex
from re import compile as re_compile


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class Game(object):

    attributes = {
        "id": str,
        "name": str,
        "key": str,
        "default": str,
        "cover": Path,
        "emulator": Emulator,
        "score": int,
        "played": int,
        "play_time": timedelta,
        "last_launch_time": timedelta,
        "last_launch_date": date,
        "installed": datetime,
        "tags": list,
        "environment": dict,
        "favorite": bool,
        "multiplayer": bool,
        "finish": bool
    }

    def __init__(self, parent, filename):
        """ Constructor

        Parameters
        ----------
        parent : gem.engine.api.GEM
            API instance
        filename : pathlib.Path
            Game file path

        Raises
        ------
        ValueError
            When the specified filename is empty or null
        FileNotFoundError
            When the specified filename not exists on filesystem
        """

        if not filename:
            raise ValueError("Cannot use an empty value as filename")

        if isinstance(filename, str):
            filename = Path(filename).expanduser()

        if not filename.exists():
            raise FileNotFoundError(f"Cannot found '{filename}' in filesytem")

        # ----------------------------------------
        #   Variables
        # ----------------------------------------

        self.__parent = parent

        self.__path = filename

        # ----------------------------------------
        #   Initialization
        # ----------------------------------------

        # Initialize attributes
        self.__init_attributes()

        # Initialize variables
        if self.__parent is not None:
            self.__init_from_database()

    def __init_attributes(self):
        """ Initialize object attributes
        """

        for key, key_type in self.attributes.items():

            if key_type is Emulator or key_type is Path:
                setattr(self, key, None)

            elif key_type is datetime and key == "installed":
                setattr(self, key, None)

            elif key_type is date:
                setattr(self, key, date(1, 1, 1))

            elif key_type is bool:
                setattr(self, key, False)

            else:
                setattr(self, key, key_type())

        setattr(self, "id", generate_identifier(self.__path))
        setattr(self, "name", self.__path.stem)
        setattr(self, "installed", get_creation_datetime(self.__path))

        self.environment.clear()
        if self.__parent is not None:
            environment = self.__parent.environment

            if self.id in environment.keys():
                for option in environment.options(self.id):
                    self.environment[option.upper()] = environment.get(
                        self.id, option, fallback=str())

    def __init_from_database(self):
        """ Initialize object with database results
        """

        # Retrieve data from database
        data = self.__parent.database.get(
            "games", {"filename": self.__path.name})

        if data is not None:

            # Retrieve time from a string (HH:MM:SS.MS)
            regex = re_compile(r"(\d+):(\d+):(\d+)[\.\d*]?")

            # Convert old column name from database to the new object scheme
            convert_keys = dict(
                play="played",
                arguments="default",
                last_play="last_launch_date",
                last_play_time="last_launch_time")

            for key, value in data.items():

                if key in convert_keys.keys():
                    key = convert_keys[key]

                if key in self.attributes.keys():
                    key_type = self.attributes[key]

                    if key_type is Path and type(value) is str:

                        path = Path(value).expanduser().resolve()
                        if len(value) == 0:
                            path = None

                        setattr(self, key, path)

                    elif key_type is Emulator and type(value) is str:
                        setattr(self, key, self.__parent.get_emulator(value))

                    elif key_type is int:
                        setattr(self, key, int(value))

                    elif key_type is bool:
                        setattr(self, key, bool(value))

                    elif key_type is list and type(value) is str:

                        if len(value) > 0:
                            value = list(set(value.strip().split(';')))
                            value.sort()

                            setattr(self, key, value)

                    elif key_type is dict:
                        pass

                    elif key_type is date:
                        day, month, year = 1, 1, 1

                        # Old GEM format
                        if len(value) > 10:
                            day, month, year = value.split()[0].split('-')

                        # ISO 8601 format
                        elif len(value) > 0:
                            year, month, day = value.split('-')

                        setattr(self, key,
                                date(int(year), int(month), int(day)))

                    elif key_type is timedelta:
                        result = regex.match(value)

                        if result is not None:
                            hours, minutes, seconds = result.groups()

                            setattr(self, key, timedelta(
                                    hours=int(hours),
                                    minutes=int(minutes),
                                    seconds=int(seconds)))

                    elif key_type is str and len(value) > 0:
                        setattr(self, key, value)

    def __str__(self):
        """ Return a formatted string when using print function
        """

        return f"[{self.id}] {self.__path}"

    def as_dict(self):
        """ Return object as dictionary structure

        Returns
        -------
        dict
            Data structure
        """

        return {
            "name": self.name,
            "filename": self.path.name,
            "favorite": self.favorite,
            "multiplayer": self.multiplayer,
            "finish": self.finish,
            "score": self.score,
            "play": self.played,
            "play_time": parse_timedelta(self.play_time),
            "last_play_time": parse_timedelta(self.last_launch_time),
            "last_play": self.last_launch_date,
            "emulator": self.emulator,
            "arguments": self.default,
            "tags": ';'.join(self.tags),
            "key": self.key,
            "cover": self.cover
        }

    def copy(self, filename):
        """ Copy game data into a new instance

        Parameters
        ----------
        filename : pathlib.Path
            Game file path

        Returns
        -------
        gem.engine.game.Game
            Game object
        """

        game = Game(self.__parent, filename)

        for key in self.attributes.keys():
            setattr(game, key, getattr(self, key, None))

        game.id = generate_identifier(filename)
        game.name = filename.stem

        return game

    def reset(self):
        """ Reset game attributes
        """

        self.__init_attributes()

    @property
    def path(self):
        """ Return the game absolute file path

        Returns
        -------
        pathlib.Path
            Game file path
        """

        return self.__path

    @property
    def extension(self):
        """ Return extension from filepath

        Returns
        -------
        str
            filename
        """

        extension = str()

        # Only retrieve extensions and not part of the name
        for subextension in self.__path.suffixes:
            if subextension not in self.__path.stem:
                extension += subextension.lower()

        return extension

    @property
    def screenshots(self):
        """ Get screenshots list
        """

        if self.emulator is not None:
            return self.emulator.get_screenshots(self)

        return list()

    @property
    def savestates(self):
        """ Get savestates list
        """

        if self.emulator is not None:
            return self.emulator.get_savestates(self)

        return list()

    def command(self, fullscreen=False):
        """ Generate a launch command

        Parameters
        ----------
        fullscreen : bool, optional
            Use fullscreen parameters (Default: False)

        Returns
        -------
        list or None
            Command launcher parameters list, None otherwise
        """

        if self.emulator is not None:
            return self.emulator.get_command_line(self, fullscreen=fullscreen)

        return None

    def update_installation_date(self):
        """ Reload installation date from game file
        """

        setattr(self, "installed", get_creation_datetime(self.__path))



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Geode
from geode_gem.widgets.common import GeodeGtkCommon

# GObject
from gi.repository import Gtk, Pango


# ------------------------------------------------------------------------------
#   Class
# ------------------------------------------------------------------------------

class CommonCellRenderer(GeodeGtkCommon):

    # Define properties which will be used with Gtk.CellRenderer.set_property
    __properties__ = dict()

    __setters__ = {
        "set_alignment": (0.5, 0.5),
        "set_padding": (0, 0),
    }

    def __init__(self, subclass, *args, **kwargs):
        """ Constructor

        Parameters
        ----------
        subclass : Gtk.MenuItem
            Subclass widget type
        title : str
            Menu item label
        """

        GeodeGtkCommon.__init__(self, subclass, **kwargs)

        # Store cells attributes which are used by GeodeGtkTreeViewColumn
        self.attributes = kwargs.get("attributes", dict())

        # Store cells properties
        self.properties = self.__properties__.copy()
        self.properties.update(kwargs.get("properties", dict()))

        for key, value in self.properties.items():
            self.set_property(key, value)


class GeodeGtkCellRendererAccel(CommonCellRenderer, Gtk.CellRendererAccel):

    __properties__ = {
        "placeholder-text": str(),
    }

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.treeview.CommonCellRenderer
        """

        CommonCellRenderer.__init__(
            self, Gtk.CellRendererAccel, *args, **kwargs)


class GeodeGtkCellRendererPixbuf(CommonCellRenderer, Gtk.CellRendererPixbuf):

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.treeview.CommonCellRenderer
        """

        CommonCellRenderer.__init__(
            self, Gtk.CellRendererPixbuf, *args, **kwargs)


class GeodeGtkCellRendererText(CommonCellRenderer, Gtk.CellRendererText):

    __properties__ = {
        "editable": False,
        "ellipsize": Pango.EllipsizeMode.NONE,
        "placeholder-text": str(),
    }

    def __init__(self, *args, **kwargs):
        """ See geode_gem.ui.widgets.treeview.CommonCellRenderer
        """

        CommonCellRenderer.__init__(
            self, Gtk.CellRendererText, *args, **kwargs)



# ------------------------------------------------------------------------------
#  Copyleft 2015-2021  PacMiam
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
# ------------------------------------------------------------------------------

# Filesystem
from pathlib import Path
from shutil import rmtree

# GEM
from geode_gem.engine.api import GEM
from geode_gem.engine.utils import copy, get_data
from geode_gem.engine.lib.configuration import Configuration

from geode_gem.ui.data import Icons, Columns, Folders, Metadata
from geode_gem.ui.utils import magic_from_file

# Logging
from logging import getLogger

# System
from argparse import ArgumentParser
from os import environ

# Translation
from gettext import textdomain, bindtextdomain
from gettext import gettext as _


# ------------------------------------------------------------------------------
#   Launcher
# ------------------------------------------------------------------------------

def copy_file(path, new_path):
    """

    Parameters
    ----------
    path : pathlib.Path
        Original file path to copy
    new_path : pathlib.Path
        Destination path

    Returns
    -------
    bool
        True is copy is successful, False, otherwise
    """

    if path.exists() and not new_path.exists():
        copy(path, new_path)
        return True

    return False


def make_directory(path, mode=0o755, make_parents=True):
    """ Generate a directory

    Parameters
    ----------
    path : pathlib.Path
        Directory path
    mode : octal integer, optional
        Directory access permissions
    make_parents : bool, optional
        Ensure to create all the directories from path if missing

    Returns
    -------
    bool
        True is creation is successful, False, otherwise
    """

    if not path.exists():
        path.mkdir(mode=mode, parents=make_parents)
        return True

    return False


def init_localization(egg_name):
    """ Initialize application localization

    Parameters
    ----------
    egg_name : str
        Python egg name
    """

    bindtextdomain(egg_name, localedir=str(get_data("data", "i18n")))
    textdomain(egg_name)


def init_environment():
    """ Initialize main environment

    Returns
    -------
    geode_gem.engine.lib.configuration.Configuration
        Metadata configuration file instance for testing purpose
    """

    # Initialize metadata
    metadata = Configuration(get_data("data", "config", "metadata.conf"))

    # Retrieve metadata informations
    if metadata.has_section("metadata"):
        for key, value in metadata.items("metadata"):
            setattr(Metadata, key.upper(), value)

    # Retrieve icons informations
    if metadata.has_section("icons"):
        for key, value in metadata.items("icons"):
            setattr(Icons, key.upper(), value)
            setattr(Icons.Symbolic, key.upper(), f"{value}-symbolic")

    if metadata.has_section("icon-sizes"):
        for key, value in metadata.items("icon-sizes"):
            setattr(Icons.Size, key.upper(), value.split())

    # Retrieve columns informations
    if metadata.has_section("misc"):
        setattr(Columns, "ORDER",
                metadata.get("misc", "columns_order", fallback=str()))

    if metadata.has_section("list"):
        for key, value in metadata.items("list"):
            setattr(Columns.List, key.upper(), int(value))

    if metadata.has_section("grid"):
        for key, value in metadata.items("grid"):
            setattr(Columns.Grid, key.upper(), int(value))

    return metadata


def init_configuration(gem):
    """ Initialize user configuration

    Parameters
    ----------
    gem : gem.engine.api.GEM
        GEM API instance
    """

    move_collection = False

    # Configuration
    for filename in ("gem.conf", "consoles.conf", "emulators.conf"):
        path = get_data("data", "config", filename)
        new_path = gem.get_config(filename)

        if copy_file(path, new_path):
            gem.logger.debug(f"Copy {path} to {new_path}")

    # Local
    for folder in ("logs", "notes"):
        if make_directory(gem.get_local(folder)):
            gem.logger.debug(f"Generate {folder} directory")

    # Cache
    for name in ("consoles", "emulators", "games"):
        for size in getattr(Icons.Size, name.upper(), list()):
            path = Folders.CACHE.joinpath(name, f"{size}x{size}")

            if make_directory(path):
                gem.logger.debug(f"Generate {path} directory")

    # Icons
    icons_path = gem.get_local("icons")

    if not icons_path.exists():
        if make_directory(icons_path):
            gem.logger.debug(f"Generate {icons_path} directory")

        move_collection = True

    # Remove older icons collections folders (GEM < 1.0)
    else:
        for folder in ("consoles", "emulators"):
            path = icons_path.joinpath(folder)

            if not path.exists():
                continue

            if path.is_dir():
                rmtree(path)

            elif path.is_symlink():
                path.unlink()

            move_collection = True

    # Copy default icons
    if move_collection:
        gem.logger.debug("Generate consoles icons folder")

        for filename in get_data("data", "icons").glob("*.png"):

            if not filename.is_file():
                continue

            # Check the file mime-type to avoid non-image file
            mime = magic_from_file(filename, mime=True)
            if mime.startswith("image/"):
                new_path = icons_path.joinpath(filename.name)

                if copy_file(filename, new_path):
                    gem.logger.debug(f"Copy {filename} to {new_path}")


def main():
    """ Main launcher
    """

    # Initialize localization
    init_localization("geode_gem")

    # Initialize environment
    init_environment()

    # ----------------------------------------
    #   Generate arguments
    # ----------------------------------------

    parser = ArgumentParser(
        epilog=Metadata.COPYLEFT,
        description=f"{Metadata.NAME} - {Metadata.VERSION}",
        conflict_handler="resolve")

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"{Metadata.NAME} {Metadata.VERSION} "
                f"({Metadata.CODE_NAME}) - {Metadata.LICENSE}",
        help="show the current version")
    parser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="launch gem with debug flag")

    parser_api = parser.add_argument_group("api arguments")
    parser_api.add_argument(
        "--cache",
        action="store",
        metavar="FOLDER",
        default=Folders.Default.CACHE,
        help=f"set cache folder (default: {Folders.Default.CACHE})")
    parser_api.add_argument(
        "--config",
        action="store",
        metavar="FOLDER",
        default=Folders.Default.CONFIG,
        help=f"set configuration folder (default: {Folders.Default.CONFIG})")
    parser_api.add_argument(
        "--local",
        action="store",
        metavar="FOLDER",
        default=Folders.Default.LOCAL,
        help=f"set data folder (default: {Folders.Default.LOCAL})")

    parser_maintenance = parser.add_argument_group("maintenance arguments")
    parser_maintenance.add_argument(
        "--clean-cache",
        action="store_true",
        help="clean icons cache directory")

    arguments = parser.parse_args()

    # ----------------------------------------
    #   Cache directory
    # ----------------------------------------

    setattr(Folders, "CACHE",
            Path(arguments.cache, "gem").expanduser().resolve())
    make_directory(Folders.CACHE)

    if Folders.CACHE.exists() and arguments.clean_cache:

        if Folders.CACHE.is_dir():
            rmtree(Folders.CACHE)

        if not Folders.CACHE.exists():
            Folders.CACHE.mkdir(mode=0o755, parents=True)

    # ----------------------------------------
    #   Launch interface
    # ----------------------------------------

    process_status = False

    try:
        gem = GEM(arguments.config, arguments.local, arguments.debug)

        # Set cache directory
        cache_path = Folders.Default.CACHE.joinpath(gem.Instance)

        # Check display settings
        if "DISPLAY" in environ and environ.get("DISPLAY"):

            if not gem.is_locked():
                # Initialize main configuration files
                init_configuration(gem)

                getLogger(gem.Instance).info(f"Start GEM with PID {gem.pid}")

                # Start splash
                from geode_gem.ui.splash import Splash
                Splash(gem)

                # Start interface
                from geode_gem.ui.interface import MainWindow
                MainWindow(gem, cache_path)

                # Remove lock
                gem.free_lock()

            else:
                getLogger(gem.Instance).critical(
                    f"GEM is already running with PID {gem.pid}")

                from geode_gem.ui.splash import Message
                message = Message(
                    _("An instance already exists"),
                    _("GEM is already running with PID %s") % gem.pid)
                message.run()

        else:
            getLogger(gem.Instance).critical(
                "Cannot launch GEM without display")

    except ImportError:
        getLogger("gem").exception("An error occur durint modules importation")
        process_status = True

    except KeyboardInterrupt:
        getLogger("gem").warning("Terminate by keyboard interrupt")
        process_status = True

    except Exception:
        getLogger("gem").exception("An error occur during execution")
        process_status = True

    # Remove lock when an error occurs
    if process_status:
        gem.free_lock()

    return process_status


if __name__ == "__main__":
    main()
